"""This functions calculates the frequency based on the input data provided. 
Currently Darshan, recorder, and traces generated with our internal tool are supported. 

call ftio -h to see list of supported arguments. 
"""

from __future__ import annotations
import time
import sys
import numpy as np
from rich.console import Group
from rich.panel import Panel
from ftio.parse.scales import Scales
from ftio.parse.extract import get_time_behavior
from ftio.freq.freq_plot_core import convert_and_plot
from ftio.freq.helper import get_mode, MyConsole, merge_results
from ftio.freq.autocorrelation import find_autocorrelation
from ftio.freq.anomaly_detection import outlier_detection
from ftio.freq.discretize import sample_data
from ftio.freq._wavelet import (
    wavelet_disc,
    plot_wave_disc,
    wavelet_cont,
    plot_wave_cont,
)  # , welch
from ftio.freq._dft import dft, prepare_plot_dfs, display_prediction, precision_dft
from ftio.prediction.unify_predictions import merge_predictions
from ftio.freq.time_window import data_in_time_window


CONSOLE = MyConsole()


def main(cmd_input: list[str], msgs=None):  # -> dict[Any, Any]:
    """Pass variables and call main_core. The extraction of the traces
    and the parsing of the arguments is done in this function.
    """
    # prepare data
    start = time.time()
    data = Scales(cmd_input, msgs)
    data.get_data()
    args = data.args
    df = get_mode(data, args.mode)
    data = get_time_behavior(df)
    CONSOLE.set(args.verbose)
    CONSOLE.print(f"\n[cyan]Data imported in:[/] {time.time() - start:.2f} s")
    CONSOLE.print(f"[cyan]Frequency Analysis:[/] {args.transformation.upper()}")
    CONSOLE.print(f"[cyan]Mode:[/] {args.mode}")

    # get prediction
    prediction, dfs = core(data, args)

    # plot and print info
    convert_and_plot(data, dfs, args)
    display_prediction(cmd_input, prediction)
    CONSOLE.print(f"[cyan]Total elapsed time:[/] {time.time()-start:.3f} s\n")

    return prediction, args


def core(data: list[dict], args) -> tuple[dict, list]:
    """ftio core function

    Args:
        data: {
            "time": time (np.ndarray),
            "bandwidth": bandwidth (np.ndarray),
            "total_bytes": total_bytes (int),
            "ranks": i (int)
            }
        args (_type_): argparse

    Returns:
        tuple[dict,list[list[float]]]: _description_
    """
    # init
    prediction = {}
    share = {}
    dfs_out = [[],[],[],[]]

    # get predictions
    for sim in data:
        # Perform frequency analysis (dft/wavelet)
        prediction_dft, dfs, share = freq_analysis(args, sim)
        # Perform autocorrelation if args.autocorrelation is true + Merge the results into a single prediction
        prediction_auto = find_autocorrelation(args, sim, share)
        # Merge results
        prediction = merge_predictions(args, prediction_dft, prediction_auto)
        # merge plots
        dfs_out = merge_results(dfs_out, dfs)

    return prediction, dfs_out


def freq_analysis(args, data: dict) -> tuple[dict, tuple[list, list, list, list], dict]:
    """Description:
    performs sampling and that frequency technique (dft, wave_cont, or wave_disc),
    followed by creating
    a dataframe with the information for plotting

    Args:
        args (argparse): see io_args.py
        data (dict[str,np.ndarray]): containing 4 fields:
            1. bandwidth (np.array): bandwidth array
            2. time (np.array): time array indicating when bandwidth time points changed
            3. total_bytes (int): total transferred bytes
            4. ranks: number of ranks that did I/O

    Raises:
        Exception: _description_

    Returns:
        dict: Containing the prediction with the following fields:
            1. dict:  Containing result of prediction including:
                "dominant_freq" (list), "conf" (np.array), "t_start" (int), "t_end" (int), "total_bytes" (int).
            2. tuple[list, list, list, list]: for plot
            3. dict: Containing sampled data including:
                b_sampled, "freq", "t_start", "t_end", "total_bytes"
    """
    #! Init
    k = 0
    share = {}
    df_out = [[], [], [], []]
    prediction = {
        "source": {args.transformation},
        "dominant_freq": [],
        "conf": [],
        "t_start": 0,
        "t_end": 0,
        "total_bytes": 0,
        "freq": 0,
        "ranks": 0,
    }
    bandwidth = data["bandwidth"] if "bandwidth" in data else np.array([])
    time_b = data["time"] if "time" in data else np.array([])
    total_bytes = data["total_bytes"] if "total_bytes" in data else 0
    ranks = data["ranks"] if "ranks" in data else 0

    #! extract relevant data
    bandwidth, time_b, text = data_in_time_window(
        args, bandwidth, time_b, total_bytes, ranks
    )

    #! Discretize signal: sample the bandwidth
    tik = time.time()
    CONSOLE.print("[cyan]Executing:[/] Discretization\n")
    b_sampled, freq, text_disc = sample_data(bandwidth, time_b, args.freq)
    CONSOLE.print(
        Panel.fit(
            text_disc,
            style="white",
            border_style="yellow",
            title="Discretization",
            title_align="left",
        )
    )
    CONSOLE.print(f"\n[cyan]Discretization finished:[/] {time.time() - tik:.3f} s")

    #! Perform transformation
    CONSOLE.print(
        f"[cyan]Executing:[/] {args.transformation.upper()} + {args.outlier}\n"
    )
    tik = time.time()

    ##! Choose Method
    if "dft" in args.transformation:
        ##? calculate DFT
        X = dft(b_sampled)
        N = len(X)
        amp = abs(X)
        freq_arr = freq * np.arange(0, N) / N
        phi = np.arctan2(X.imag, X.real)
        conf = np.zeros(len(amp))
        # welch(bandwidth,freq)

        ##? Find dominant frequency
        (
            dominant_index,
            conf[1 : int(len(amp) / 2) + 1],
            outlier_text,
        ) = outlier_detection(amp, freq_arr, args)

        ##? ignore DC offset
        conf[0] = np.inf
        if len(amp) % 2 == 0:
            conf[int(len(amp) / 2) + 1 :] = np.flip(conf[1 : int(len(amp) / 2)])
        else:
            conf[int(len(amp) / 2) + 1 :] = np.flip(conf[1 : int(len(amp) / 2) + 1])

        ##? Assign data
        prediction["dominant_freq"] = freq_arr[dominant_index]
        prediction["conf"]          = conf[dominant_index]
        prediction["amp"]           = amp[dominant_index]
        prediction["phi"]           = phi[dominant_index]
        prediction["t_start"]       = time_b[0]
        prediction["t_end"]         = time_b[-1]
        prediction["freq"]          = freq
        prediction["ranks"]         = ranks
        prediction["total_bytes"]   = total_bytes

        #? save up to n_freq from the top candidates
        if args.n_freq > 0:
            #TODO use amp instead of conf
            arr = amp[0:int(np.ceil(len(amp)/2))]
            top_candidates = np.argsort(-arr) # from max to min
            n_freq = int(min(len(arr),args.n_freq))
            prediction["top_freq"] = {
                "freq": freq_arr[top_candidates[0:n_freq]],
                "conf": conf[top_candidates[0:n_freq]],
                "amp":  amp[top_candidates[0:n_freq]],
                "phi":  phi[top_candidates[0:n_freq]]
            }
            
        if args.autocorrelation:
            share["b_sampled"]   = b_sampled
            share["freq"]        = freq
            share["t_start"]     = prediction["t_start"]
            share["t_end"]       = prediction["t_end"]
            share["total_bytes"] = prediction["total_bytes"]

        precision_text = ""
        # precision_text = precision_dft(
        #     amp, phi, dominant_index, b_sampled, time_b[0] + np.arange(0, N) * 1 / freq, freq_arr, args.engine
        # )

        text = Group(text, outlier_text, precision_text[:-1])

        if any(x in args.engine for x in ["mat", "plot"]):
            df_out = prepare_plot_dfs(
                k,
                freq,
                freq_arr,
                conf,
                dominant_index,
                amp,
                phi,
                b_sampled,
                time_b,
                ranks,
                bandwidth,
            )
        k += 1

    elif "wave_disc" in args.transformation:
        # discrete wavelet decomposition:
        # https://edisciplinas.usp.br/pluginfile.php/4452162/mod_resource/content/1/V1-Parte%20de%20Slides%20de%20p%C3%B3sgrad%20PSI5880_PDF4%20em%20Wavelets%20-%202010%20-%20Rede_AIASYB2.pdf
        # https://www.youtube.com/watch?v=hAQQwvKsWCY&ab_channel=NathanKutz
        print("    '-> \033[1;32mPerforming discrete wavelet decomposition\033[1;0m")
        wavelet = "db1"  # dmey might be better https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html
        # wavelet = 'haar' # dmey might be better https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html
        coffs = wavelet_disc(b_sampled, wavelet, args.level)
        cc, f = plot_wave_disc(
            b_sampled, coffs, time_b, args.freq, args.level, wavelet, bandwidth
        )
        for fig in f:
            fig.show()
        cont = input("\nContinue with the DFT? [y/n]")

        if len(cont) == 0 or "y" in cont.lower():
            args.transformation = "dft"
            n = len(b_sampled)
            tmp = {
                "time": time_b[0] + 1 / freq * np.arange(0, n),
                "bandwidth": cc[0],
                "total_bytes": total_bytes,
                "ranks": ranks,
            }
            freq_analysis(args, tmp)

        sys.exit()

    elif "wave_cont" in args.transformation:
        # Continuous wavelets
        print("    '-> \033[1;32mPerforming discrete wavelet decomposition\033[1;0m")
        wavelet = "morl"
        # wavelet = 'cmor'
        # wavelet = 'mexh'
        [coefficients, frequencies, scale] = wavelet_cont(
            b_sampled, wavelet, args.level, args.freq
        )
        fig = plot_wave_cont(
            b_sampled, frequencies, args.freq, scale, time_b, coefficients
        )
        sys.exit()

    else:
        raise Exception("Unsupported decomposition specified")
    CONSOLE.print(
        Panel.fit(
            text,
            style="white",
            border_style="cyan",
            title=args.transformation.upper(),
            title_align="left",
        )
    )
    CONSOLE.print(
        f"\n[cyan]{args.transformation.upper()} + {args.outlier} finished:[/] {time.time() - tik:.3f} s"
    )
    return prediction, df_out, share


def run():
    _ = main(sys.argv)


if __name__ == "__main__":
    _ = main(sys.argv)



import pandas as pd

import ftio.plot.dash_files.constants.io_mode as io_mode


class FileData:
    def __init__(
        self,
        run: int,
        rank: int,
        name: str,
        data_actual: list[pd.DataFrame],
        data_required: list[pd.DataFrame],
    ):
        self._run = run
        self._rank = rank
        self._name = name
        self._data_actual = data_actual
        self._data_required = data_required
        self._init_masks()

    def _init_masks(self):
        self._mask = self._data_actual[1]["number_of_ranks"].isin([self.rank])
        self._mask_ind = self._data_actual[3]["number_of_ranks"].isin([self.rank])
        self._mask2 = self._data_actual[1]["file_index"][self._mask].isin([self.run])
        self._mask2_ind = self._data_actual[3]["file_index"][self._mask_ind].isin([self.run])

    @property
    def run(self) -> int:
        return self._run

    @property
    def rank(self) -> int:
        return self._rank

    @property
    def name(self) -> str:
        return self._name

    @property
    def data_actual_is_not_empty(self) -> bool:
        return len(self._data_actual) != 0

    @property
    def data_required_is_not_empty(self) -> bool:
        return len(self._data_required) != 0

    @property
    def actual_time_overlap(self):
        return self._data_actual[1]["t_overlap"][self._mask][self._mask2]

    @property
    def required_time_overlap(self):
        return self._data_required[1]["t_overlap"][self._mask][self._mask2]

    @property
    def actual_time_overlap_individual(self):
        return self._data_actual[3]["t_overlap_ind"][self._mask_ind][self._mask2_ind]

    @property
    def required_time_overlap_individual(self):
        return self._data_required[3]["t_overlap_ind"][self._mask_ind][self._mask2_ind]

    @property
    def actual_bandwidth_overlap_average(self):
        return self._data_actual[1]["b_overlap_avr"][self._mask][self._mask2]

    @property
    def required_bandwidth_overlap_average(self):
        return self._data_required[1]["b_overlap_avr"][self._mask][self._mask2]

    @property
    def actual_bandwidth_overlap_sum(self):
        return self._data_actual[1]["b_overlap_sum"][self._mask][self._mask2]

    @property
    def required_bandwidth_overlap_sum(self):
        return self._data_required[1]["b_overlap_sum"][self._mask][self._mask2]

    @property
    def actual_bandwidth_overlap_individual(self):
        return self._data_actual[3]["b_overlap_ind"][self._mask_ind][self._mask2_ind]

    @property
    def required_bandwidth_overlap_individual(self):
        return self._data_required[3]["b_overlap_ind"][self._mask_ind][self._mask2_ind]


class DataSource:
    def __init__(self, plot_core, io_mode: str):
        self._io_mode = io_mode
        self._plot_core = plot_core

        self._data = plot_core.data

        self._init_data_actual_and_required()
        self._init_names_and_ranks()
        self._init_file_data_dictionary()

    def _init_data_actual_and_required(self):
        match self.io_mode:
            case io_mode.ASYNC_READ:
                self._data_actual: list[pd.DataFrame] = (
                    self._data.df_rat
                    if not (self._data.df_rat is None or self._data.df_rat[1].empty)
                    else []
                )
                self._data_required: list[pd.DataFrame] = (
                    self._data.df_rab
                    if not (self._data.df_rab is None or self._data.df_rab[1].empty)
                    else []
                )
            case io_mode.ASYNC_WRITE:
                self._data_actual: list[pd.DataFrame] = (
                    self._data.df_wat
                    if not (self._data.df_wat is None or self._data.df_wat[1].empty)
                    else []
                )
                self._data_required: list[pd.DataFrame] = (
                    self._data.df_wab
                    if not (self._data.df_wab is None or self._data.df_wab[1].empty)
                    else []
                )
            case io_mode.SYNC_READ:
                self._data_actual: list[pd.DataFrame] = (
                    self._data.df_rst
                    if not (self._data.df_rst is None or self._data.df_rst[1].empty)
                    else []
                )
                self._data_required: list[pd.DataFrame] = []
            case io_mode.SYNC_WRITE:
                self._data_actual: list[pd.DataFrame] = (
                    self._data.df_wst
                    if not (self._data.df_wst is None or self._data.df_wst[1].empty)
                    else []
                )
                self._data_required: list[pd.DataFrame] = []
            case _:
                raise Exception("invalid mode")

    def _init_names_and_ranks(self) -> None:
        self._ranks: pd.Series = pd.Series()
        if len(self._data_actual) != 0:
            self._ranks = self._data_actual[0]["number_of_ranks"].astype(int)
        elif len(self._data_required) != 0:
            self._ranks = self._data_required[0]["number_of_ranks"].astype(int)
        names_and_ranks_df = pd.concat(
            [pd.Series(self._plot_core.names, name="filenames"), self._ranks], axis=1
        )
        self._filenames = names_and_ranks_df["filenames"].to_list()
        self._ranks = names_and_ranks_df["number_of_ranks"].to_list()
        self._ranks_unique = pd.unique(self.ranks)
        self._ranks_unique.sort()

    def _init_file_data_dictionary(self):
        self._file_data_by_filename: dict[str, FileData] = dict()
        for idx, filename in enumerate(self._filenames):
            self._file_data_by_filename[filename] = FileData(
                idx,
                self._ranks[idx],
                self._filenames[idx],
                self._data_actual,
                self._data_required,
            )

    @property
    def io_mode(self) -> str:
        return self._io_mode

    @property
    def filenames(self) -> list[str]:
        return self._filenames

    @property
    def ranks(self) -> list[int]:
        return self._ranks

    @property
    def ranks_unique(self) -> list[int]:
        return self._ranks_unique

    @property
    def fontfamily(self) -> str:
        return "Courier New, monospace"

    @property
    def fontsize(self) -> int:
        return 17

    @property
    def width_figure(self) -> int:
        return self._plot_core.width

    @property
    def height_figure(self) -> int:
        return self._plot_core.height

    @property
    def file_data_by_filename(self) -> dict[str, FileData]:
        return self._file_data_by_filename

    @property
    def individual_is_selected(self) -> bool:
        return self._data.args.ind

    @property
    def n_shown_samples(self) -> int:
        return self._data.args.n_shown_samples

    @property
    def merge_plots_is_selected(self) -> bool:
        return self._data.args.merge_plots


def get_data_source(plot_core, io_mode_: str) -> DataSource:
    """Creates and returns a suitable DataSource-class depending on the specified io_mode.

    Args:
        plot_core (plot_core): plot_core-class
        mode (str): input-output-mode
            current valid modes:
                io_mode.ASYNC_READ,
                io_mode.ASYNC_WRITE,
                io_mode.SYNC_READ,
                io_mode.SYNC_WRITE

    Raises:
        Exception: raises if mode can't be handled

    Returns:
        DataSource: Suitable DataSource-class for the mode
    """
    match io_mode_:
        case io_mode.ASYNC_READ:
            return DataSource(plot_core, io_mode.ASYNC_READ)
        case io_mode.ASYNC_WRITE:
            return DataSource(plot_core, io_mode.ASYNC_WRITE)
        case io_mode.SYNC_READ:
            return DataSource(plot_core, io_mode.SYNC_READ)
        case io_mode.SYNC_WRITE:
            return DataSource(plot_core, io_mode.SYNC_WRITE)
        case _:
            raise Exception("invalid mode")



def convert()->tuple[dict[str,str],dict[str,tuple[str,float]]]:
    """Converts input according to pattern into and ftio supported file format.
    The translations dictionary species the matching fields from ftio/parse/sample.py
    Returns:
        tuple[dict[str,str],dict[str,tuple[str,float]]]: pattern and translation
            (1) pattern (dict[str,str]): dictionary containing name and a regex 
                        expression to find the custom pattern.
            (2) translate (dict[str,tuple[str,(optional)float]]): dictionary 
                        containing matching filed from sample.py and the matching 
                        the name from the pattern. The unit can be optionally specified
    """
    pattern = {
        "avg_thruput_mib": r"avg_thruput_mib:\s*\[([\d.\d,\s*]+)\]",
        "end_t_micro": r"end_t_micro:\s*\[([\d,\s*]+)\]",
        "start_t_micro": r"start_t_micro:\s*\[([\d,\s*]+)\]",
        # "req_size": r"req_size:\s+\[([\d,\s]+)\]",
        "total_bytes": r"total_bytes:\s*+(\d+)",
        "total_iops": r"total_iops:\s*+(\d+)",
        }

    # Define map according to sample.py class, along with the scale if any:
    # ftio_field: ("custom_name", scale)
    # ftio unit are default in bytes, b/s, ...
    # scale applies ftio_field = custom_name*scale
    translate = {
        "bandwidth": {
            "b_rank_avr": ("avg_thruput_mib",1.07*1e+6),
            "t_rank_e": ("end_t_micro", 1e-3),
            "t_rank_s": ("start_t_micro", 1e-3)
            },
        # "max_bytes_per_phase": "req_size",
        "total_bytes": "total_bytes",
        "max_io_ops_per_rank": "total_iops"
        }
    return pattern, translate


"""Helper function for frequency techniques    
"""

import plotly.graph_objects as go
from rich.console import Console


def get_mode(data, mode):
    """used after get_data() to extract df. The df
    contains all sims group by mode

    Args:
        data (Scales): simulation data
        mode (str): "read_sync", "write_sync", "read_async" or "write_async"

    Raises:
        Exception: unsupported mode

    Returns:
        df: pandas dataframe containing data group by mode
    """
    mode = mode.lower()
    if "read" in mode:
        if "async" in mode:
            return data.df_rat
        elif "sync" in mode:
            return data.df_rst
    if "write" in mode:
        if "async" in mode:
            return data.df_wat
        elif "sync" in mode:
            return data.df_wst
    raise Exception("undefined mode set")


def get_sim(data, mode):
    mode = mode.lower()
    if "read" in mode:
        if "async" in mode:
            return data.read_async_t
        elif "sync" in mode:
            return data.read_sync
    if "write" in mode:
        if "async" in mode:
            return data.write_async_t
        elif "sync" in mode:
            return data.write_sync
    raise Exception("undefined mode set")



def format_plot(fig) -> go.Figure:
    """makes plots uniform

    Args:
        fig (pltoly figure)
    """
    fig.update_layout(
        plot_bgcolor="white",
        legend=dict(
            bgcolor="rgba(255,255,255,.99)",
            bordercolor="Black",
            borderwidth=1,
        ),
        font=dict(family="Courier New, monospace", size=24, color="black"),
        # margin=dict(l=5, r=5, t=5, b=5) #IEEE
        margin=dict(t=25),
    )

    fig.update_xaxes(
        ticks="outside",
        # tickwidth=1,
        ticklen=10,
        showgrid=True,
        # gridwidth=1,
        mirror=True,
        showline=True,
        linecolor="black",
        gridcolor="lightgrey",
        minor_ticks="outside",
        minor=dict(ticklen=2),
    )

    fig.update_yaxes(
        ticks="outside",
        # tickwidth=1,
        ticklen=10,
        showgrid=True,
        # gridwidth=1,
        mirror=True,
        showline=True,
        linecolor="black",
        gridcolor="lightgrey",
        minor_ticks="outside",
        minor=dict(ticklen=2),
    )

    return fig


def merge_results(
    df0: tuple[list, list, list, list], df1: tuple[list, list, list, list]
) -> tuple[list, list, list, list]:
    """merges results for several files

    Args:
        df0 (list): _description_
        df1 (list): _description_

    Returns:
        list: _description_
    """
    for i in range(0, len(df0)):
        df0[i].extend(df1[i])

    return df0


class MyConsole(Console):
    """Console child class that overwrites
    the print method for silent version

    Args:
        Console (_type_): _description_
    """

    def __init__(self):
        super().__init__()
        self.verbose = False

    def set(self, flag):
        if flag:
            self.verbose = True
        else:
            self.verbose = False

    def print(self, s):
        if self.verbose:
            Console.print(self, s)

    def info(self, s):
        Console.print(self, s)



"""Input/output modes"""

ASYNC_READ: str = "read_async"
ASYNC_WRITE: str = "write_async"
SYNC_READ: str = "read_sync"
SYNC_WRITE: str = "write_sync"
TIME: str = "time"

ALL_MODES: list[str] = [ASYNC_WRITE, ASYNC_READ, SYNC_WRITE, SYNC_READ, TIME]

MODE_STRING_BY_MODE: dict[str, str] = {
    ASYNC_READ: "Async read",
    ASYNC_WRITE: "Async write",
    SYNC_READ: "Sync read",
    SYNC_WRITE: "Sync write",
    TIME: "Time",
}



import os
from sys import platform
from threading import Thread
import plotly.offline
from rich.console import Console

def create_html(figs:list,render :str,configuration:dict,name:str="freq") -> None:
    if platform == "linux" or platform == "linux2":
        os.system(f"rm -f ./{name}.html || true")
        os.system("rm -rf io_anomality_freq_images || true ")

    if render == "dynamic":
        template = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
        <meta charset="utf-8"/>
        <script>{plotly}</script>
        </head>
        <body>
        {plots}
        </body>
        </html>
        """
        s = ""
        for fig in figs:
            # s = s + fig.to_html(include_plotlyjs=False) + "\n"
            s= s + fig.to_html(config=configuration,include_plotlyjs=False) + "\n"
        template = template.format(
            plotly=plotly.offline.get_plotlyjs(), plots=s
        )
        with open(f"{name}.html", "a") as file:
            file.write(template)
        os.system(f"open ./{name}.html &\n")
    else:
        os.mkdir("io_predicition_anomality_images")
        # extension='jpg'
        extension='svg'
        # extension = "png"
        print("-> Generating %s figures" % extension.upper())
        threads = []
        length = len(figs)
        plotly.io.json.config.default_engine = "orjson"
        for fig in figs:
            index = figs.index(fig)
            threads.append(
                Thread(
                    target=create_static_figure,
                    args=(fig, index, length, extension),
                )
            )

        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        os.system("open . &")


def create_static_figure(fig, index:int, length:int, extension:str)-> None:
    """Creates static HTML figures

    Args:
        fig (plotly figure): _description_
        index (int): index of the figure
        length (int): total figures
        extension (str): desired extension
    """
    console = Console()
    scale = 1
    console.print("working on [cyan]figure (%i/%i) [/]" % (index + 1, length))
    if fig.layout.title.text:
        file_name = f"io_predicition_freq_images/{fig.layout.title.text.replace(' ', '_').replace('(', '_').replace(')', '_')}.{extension}"
    else:
        file_name = f"io_predicition_freq_images/{index}.{extension}"

    if "svg" not in extension and "pdf" not in extension:
        scale = 5

    fig.write_image(file_name, scale=scale)
    if index == 0:
            os.system(f"open {file_name} || true ")

    console.print(f"[cyan]figure ({index + 1}/{length}) [/]created")




"""Performs prediction with Pools (ProcessPoolExecutor) and a callback mechanism"""
from __future__ import annotations
from concurrent.futures import ProcessPoolExecutor
import ftio.prediction.monitor as pm
from ftio.prediction.probability_analysis import find_probability
from ftio.prediction.helper import  print_data
from ftio.prediction.analysis import ftio_process
# from ftio.prediction.async_process import handle_in_process


def predictor_with_pools(filename, data, queue, count, hits, start_time, aggregated_bytes, args):
    """performs prediction in ProcessPoolExecuter. FTIO is a submitted future and probability is calculated as a callback

    Args:
        filename (str): name of file
        data (Manager().list): List of dicts with all predictions so far
        queue (Manager().Queue): queue for FTIO data
        count (Manager().Value): number of prediction
        hits (Manager().Value): hits indicating how often a dominant frequency was found
        start_time (Manager().Value): start time window for ftio
        aggregated_bytes (Manager().Value): total bytes transferred so far
        args (list[str]): additional arguments passed to ftio
    """
    # Init: Monitor a file
    stamp, _ = pm.monitor(filename,"")

    # Loop and predict if changes occur
    try:
        while True:
            with ProcessPoolExecutor(max_workers=1) as executor:
                # monitor
                stamp, _ = pm.monitor(filename, stamp)
                future = executor.submit(ftio_future, data, queue, count, hits, start_time, aggregated_bytes, args)
                future.add_done_callback(probability_callback)
    except KeyboardInterrupt:
        print_data(data)
        print("-- done -- ")


def ftio_future(data, queue , count, hits, start_time, aggregated_bytes, args: list[str]) -> list[dict]:
    """Performs prediction made up of two part: (1) Executes FTIO and (2) appends to data the value

    Args:
        data (Manager().list): List of dicts with all predictions so far
        queue (Manager().Queue): queue for FTIO data
        count (Manager().Value): number of prediction
        hits (Manager().Value): hits indicating how often a dominant frequency was found
        start_time (Manager().Value): start time window for ftio
        aggregated_bytes (Manager().Value): total bytes transferred so far
        args (list[str]): additional arguments passed to ftio
    """
    ftio_process(queue, count, hits, start_time, aggregated_bytes, args)
    while not queue.empty():
        data.append(queue.get())
    return data


def probability_callback(future):
    """executes the probability calculation in a callback

    Args:
        future (Future): _description_
    """
    _ = find_probability(future.result()) #the queue






import numpy as np
from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
from ftio.freq.freq_plot_core import convert_and_plot
from ftio.parse.bandwidth import overlap


def test_api():
    ranks = 10
    total_bytes = 100
    b_rank = [
        0.0,
        0.0,
        1000.0,
        1000.0,
        0.0,
        0.0,
        1000.0,
        1000.0,
        0.0,
        0.0,
        1000.0,
        1000.0,
        0.0,
        0.0,
    ]
    t_rank_s = [
        0.5,
        0.0,
        10.5,
        10.0,
        20.5,
        20.0,
        30.5,
        30.0,
        40.5,
        40.0,
        50.5,
        50.0,
        60.5,
        60,
    ]
    t_rank_e = [
        5.0,
        4.5,
        15.0,
        14.5,
        25.0,
        24.5,
        35.0,
        34.5,
        45.0,
        44.5,
        55.0,
        54.5,
        65.0,
        64.5,
    ]
    b, t = overlap(b_rank, t_rank_s, t_rank_e)
    # command line arguments
    argv = ["-e", "no"]  # ["-e", "mat"]
    # set up data
    data = {
        "time": np.array(t),
        "bandwidth": np.array(b),
        "total_bytes": total_bytes,
        "ranks": ranks,
    }
    # parse args
    args = parse_args(argv, "ftio")
    # perform prediction
    prediction, dfs = core([data], args)
    # plot and print info
    convert_and_plot(data, dfs, args)
    display_prediction("ftio", prediction)
    assert True



"""Converts results from TMIO / Darshan files into Extra-P supported format
"""
import numpy as np
from ftio.parse.scales import Scales
from ftio.parse.helper import scale_metric


class Print:
    def __init__(self, args):
        self.data = Scales(args)
        self.data.get_data()
        self.args = self.data.args

    def print_txt(self):
        self.file = open("./scale.txt", "w")
        self.file.write("pARAMETER x\n\n")
        self.print_points()
        self.print_regions_txt()
        self.file.write("\n")
        self.file.close()
        print("\033[1;32m------------------- done -------------------\n\033[1;0m")

    def print_json_lines(self):
        self.file = open("./scale.jsonl", "w")
        self.print_regions_jsonl()
        self.file.close()
        print("\033[1;32m------------------- done -------------------\n\033[1;0m")

    def print_regions_jsonl(self):
        self.print_io_read_sync("jsonl")
        self.print_io_write_sync("jsonl")
        self.print_io_read_async_t("jsonl")
        self.print_io_read_async_b("jsonl")
        self.print_io_write_async_t("jsonl")
        self.print_io_write_async_b("jsonl")
        self.print_io_time("jsonl")
        self.print_io_percent("jsonl")

    def print_regions_txt(self):
        self.print_io_read_sync()
        self.print_io_write_sync()
        self.print_io_read_async_t()
        self.print_io_read_async_b()
        self.print_io_write_async_t()
        self.print_io_write_async_b()
        self.print_io_time()
        self.print_io_percent()

    def print_io_read_sync(self, type="txt"):
        self.print_io_mode("read_sync", type)

    def print_io_write_sync(self, type="txt"):
        self.print_io_mode("write_sync", type)

    def print_io_read_async_t(self, type="txt"):
        self.print_io_mode("read_async_t", type)

    def print_io_read_async_b(self, type="txt"):
        self.print_io_mode("read_async_b", type)

    def print_io_write_async_t(self, type="txt"):
        self.print_io_mode("write_async_t", type)

    def print_io_write_async_b(self, type="txt"):
        self.print_io_mode("write_async_b", type)

    def print_io_mode(self, mode, type):
        self.print_data(mode, "total_bytes", f"{mode}->total_bytes", "Size (B)", type)
        self.print_data(
            mode, "max_bytes_per_rank", f"{mode}->max_bytes_per_rank", "Size (B)", type
        )
        self.print_data(
            mode,
            "max_bytes_per_phase",
            f"{mode}->max_bytes_per_phase",
            "Size (B)",
            type,
        )
        self.print_data(
            mode,
            "max_io_phases_per_rank",
            f"{mode}->max_io_phases_per_rank",
            "Hits",
            type,
        )
        self.print_data(
            mode, "total_io_phases", f"{mode}->total_io_phases", "Hits",
            type,
        )
        self.print_data(
            mode,
            "max_io_ops_in_phase",
            f"{mode}->max_io_ops_in_phase",
            "Hits",
            type,
        )
        self.print_data(
            mode,
            "max_io_ops_per_rank",
            f"{mode}->max_io_ops_per_rank",
            "Hits",
            type,
        )
        self.print_data(mode, "total_io_ops", f"{mode}->total_io_ops", "Hits", type)
        self.print_data(mode, "number_of_ranks", f"{mode}->number_of_ranks", "Hits", type)
        self.print_data(mode, "bandwidth.app", f"{mode}->app", print_type=type)
        self.print_data(mode, "bandwidth.appH", f"{mode}->appH", print_type=type)
        # self.print_data(mode, 'bandwidth.b',                      f"{mode}->per_rank->b"                   , print_type = type)
        # self.print_data(mode, 'bandwidth.b_overlap_avr',          f"{mode}->per_rank->b_overlap_avr"         , print_type = type)
        # self.print_data(mode, 'bandwidth.b_overlap_sum',          f"{mode}->per_rank->b_overlap_sum"         , print_type = type)
        if "b" in mode:
            self.print_data(mode, "bandwidth.app_ind", f"{mode}->B_E", print_type=type)
            self.print_data(mode, "bandwidth.app_avr", f"{mode}->B_A", print_type=type)
            self.print_data(mode, "bandwidth.app_sum", f"{mode}->B_S", print_type=type)
        else:
            self.print_data(mode, "bandwidth.app_ind", f"{mode}->T_E", print_type=type)
            self.print_data(mode, "bandwidth.app_avr", f"{mode}->T_A", print_type=type)
            self.print_data(mode, "bandwidth.app_sum", f"{mode}->T_S", print_type=type)
        self.print_data(
            mode, "bandwidth.min", f"{mode}->per_rank->min", print_type=type
        )
        self.print_data(
            mode, "bandwidth.max", f"{mode}->per_rank->max", print_type=type
        )
        self.print_data(
            mode, "bandwidth.median", f"{mode}->per_rank->median", print_type=type
        )
        self.print_data(
            mode,
            "bandwidth.weighted_harmonic_mean",
            f"{mode}->per_rank->weighted_harmonic_mean",
            print_type=type,
        )
        self.print_data(
            mode,
            "bandwidth.harmonic_mean",
            f"{mode}->per_rank->harmonic_mean",
            print_type=type,
        )
        self.print_data(
            mode,
            "bandwidth.arithmetic_mean",
            f"{mode}->per_rank->arithmetic_mean",
            print_type=type,
        )

    def print_io_time(self, type="txt"):
        try:
            self.print_data(
                "io_time",
                "delta_t_total",
                "io_time->total_time",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_overhead",
                "io_time->total_time->lib_overhead_time",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_overhead_post_runtime",
                "io_time->total_time->lib_overhead_time->delta_t_overhead_post_runtime",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_overhead_peri_runtime",
                "io_time->total_time->lib_overhead_time->delta_t_overhead_peri_runtime",
                "Time (s)",
                print_type=type,
            )

            self.print_data(
                "io_time",
                "delta_t_agg",
                "io_time->total_time->total_app_time",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_com",
                "io_time->total_time->total_app_time->total_com_time",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_agg_io",
                "io_time->total_time->total_app_time->total_io_time",
                "Time (s)",
                print_type=type,
            )

            self.print_data(
                "io_time",
                "delta_t_sr",
                "io_time->total_time->total_app_time->total_io_time->read_sync",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_ar_lost",
                "io_time->total_time->total_app_time->total_io_time->delta_t_ar_lost",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_sw",
                "io_time->total_time->total_app_time->total_io_time->write_sync",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_aw_lost",
                "io_time->total_time->total_app_time->total_io_time->delta_t_aw_lost",
                "Time (s)",
                print_type=type,
            )
            # self.print_data('io_time', 'delta_t_sr',      'io_time->read_sync',       'time', print_type = type)
            self.print_data(
                "io_time",
                "delta_t_ara",
                "io_time->read_async_t",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_arr",
                "io_time->read_async_b",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_ar_lost",
                "io_time->delta_t_ar_lost",
                "Time (s)",
                print_type=type,
            )

            # self.print_data('io_time', 'delta_t_sw',      'io_time->write_sync',      'time', print_type = type)
            self.print_data(
                "io_time",
                "delta_t_awa",
                "io_time->write_async_t",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_awr",
                "io_time->write_async_b",
                "Time (s)",
                print_type=type,
            )
            self.print_data(
                "io_time",
                "delta_t_aw_lost",
                "io_time->delta_t_aw_lost",
                "Time (s)",
                print_type=type,
            )

        except:
            pass

    def print_io_percent(self, type="txt"):
        try:
            self.print_data(
                "io_percent",
                "TI",
                "io_percent-> I/O to Total",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TC",
                "io_percent-> Compute to Total",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CI",
                "io_percent-> I/O to Compute",
                "Ratio",
                print_type=type,
            )

            self.print_data(
                "io_percent",
                "TAWB",
                "io_percent->AW->AW to Total (B)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TAWT",
                "io_percent->AW->AW to Total (T)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TAWD",
                "io_percent->AW->AW to Total (D)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "IAWB",
                "io_percent->AW->AW to IO (B)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "IAWT",
                "io_percent->AW->AW to IO (T)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "IAWD",
                "io_percent->AW->AW to IO (D)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CAWB",
                "io_percent->AW->AW to Comput (B)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CAWT",
                "io_percent->AW->AW to Comput (T)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CAWD",
                "io_percent->AW->AW to Comput (D)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TARB",
                "io_percent->AR->AR to Total (B)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TART",
                "io_percent->AR->AR to Total (T)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TARD",
                "io_percent->AR->AR to Total (D)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "IARB",
                "io_percent->AR->AR to IO (B)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "IART",
                "io_percent->AR->AR to IO (T)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "IARD",
                "io_percent->AR->AR to IO (D)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CARB",
                "io_percent->AR->AR to Compute (B)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CART",
                "io_percent->AR->AR to Compute (T)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CARD",
                "io_percent->AR->AR to Compute (D)",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TSW",
                "io_percent->SW->SW to Total",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "ISW",
                "io_percent->SW->SW to IO",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CSW",
                "io_percent->SW->SW to Compute",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TSR",
                "io_percent->SR-> SR to Total",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "ISR",
                "io_percent->SR-> SR to IO",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CSR",
                "io_percent->SR-> SR to Compute",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "TO",
                "io_percent->overhead-> lib overhead to Total",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "IO",
                "io_percent->overhead-> lib overhead to IO",
                "Ratio",
                print_type=type,
            )
            self.print_data(
                "io_percent",
                "CO",
                "io_percent->overhead-> lib overhead to Compute",
                "Ratio",
                print_type=type,
            )

        except:
            pass

    def print_data(
        self,
        io_mode,
        var="bandwidth.app",
        call_path="",
        metric="Bandwidth (B/s)",
        print_type="txt",
    ):
        if self.check_non_empty(io_mode, var):
            if not call_path:
                call_path = io_mode
            if "txt" in print_type:  # txt file
                self.file.write("\nREGION %s\nMETRIC %s\n" % (call_path, metric))
                for i in range(0, self.data.n):
                    value = getattr(self.data.s[i], io_mode)
                    if "bandwidth" in var:
                        art = getattr(value.bandwidth, var[10:])
                    else:
                        art = getattr(value, var)
                    if isinstance(art, list):
                        self.file.write("DATA ")
                        for i, _ in enumerate(art):
                            self.file.write(f"{art[i]:e}")
                        self.file.write("\n")
                    else:
                        self.file.write(f"DATA {art:e} \n")
                self.file.write("\n")
            elif "jsonl" in print_type:
                order = 1
                for i in range(0, self.data.n):
                    value = getattr(self.data.s[i], io_mode)
                    if "bandwidth" in var:
                        art = getattr(value.bandwidth, var[10:])
                    else:
                        art = getattr(value, var)
                        art = -1 if np.isnan(art) else art
                    if isinstance(art, list):
                        for j, _ in enumerate(art):
                            if j == 0:
                                if self.args.scale:
                                    metric, order = scale_metric(metric,art[j])
                            self.file.write(
                                f'{{"params":{{"Processes":{self.data.s[i].ranks}}},"callpath":"{call_path}","metric":"{metric}","value":{art[j]*order:e} }}\n'
                            )
                    else:
                        if self.args.scale:
                            metric, order = scale_metric(metric,art)
                        self.file.write(
                            f'{{"params":{{"Processes":{self.data.s[i].ranks}}},"callpath":"{call_path}","metric":"{metric}","value":{art*order:e} }}\n'
                        )
            else:
                pass

    def check_non_empty(self, io_mode, var):
        if self.data.n < 1:
            return False
        else:
            value = getattr(self.data.s[0], io_mode)
            if "bandwidth" in var:
                if hasattr(value.bandwidth, var[10:]):
                    art = getattr(value.bandwidth, var[10:])
                else:
                    return False
            else:
                try:
                    art = getattr(value, var)
                except AttributeError:
                    return False

            if isinstance(art, list) and not art:
                return False
            else:
                return True

    def print_points(self):
        for run in self.data.s:
            run.print_rank(self.file)



"""Generates HTML page out of plotly images
    Used by plot_core.py
"""
import os
import platform as plat
from sys import platform
import plotly.offline
from threading import Lock




class print_html:

    def __init__(self, args, path, names, filename = "main.html", outdir = "io_results") -> None:
        """generates HTML report
        Args:
            filename (str): main html file name
            args (list): contains render (static or dynamic) 
                        and show (True or False) flag
            path (str): output location
            outdir (str): output dir name
        """
        self.render = args.render
        self.path = path
        self.names = names
        self.filename = filename
        self.outdir = outdir
        self.show = not args.no_disp
        self.lock = Lock()

    #! ----------------------- Plot to HTML ------------------------------
    #**********************************************************************
    #*                       1. generate_html
    #**********************************************************************
    def generate_html_start(self) -> None:
        
        if not "stat" in self.render:  # only generate a single image
            pwd = os.getcwd()
            if len(self.path) <= 1 and not any(ext in self.path[0] for ext in ["json","darshan","msgpack", "txt"]):
                pwd = os.path.join(pwd,os.path.relpath(self.path[0],pwd))

            self.path = os.path.join(pwd,self.outdir)
            print(f"\n\033[1;34mGenerating HTML report \033[1;0m\n '->Directory is: {self.path}")
            if not os.path.exists(self.path):
                os.mkdir(self.path)
                print(f" '-> Directory {self.outdir} created in {pwd}")
            else:
                print(f" '-> Directory {self.outdir}  exists  in {pwd}")
        
            #?* create main HTML file
            #*-------------------------
            # html_files = ["write_async.html","read_async.html","write_sync.html","read_sync.html","time.html"] 
            print(f"\n '-> To see intermediate result call: \n     open {self.path}/main.html \n")
            self.filename = os.path.join(self.path,self.filename)
            print(" '-> Generated main.html ")
            with open(self.filename, "w") as file:
                file.write("<html><head></head><body>\n")
                file.write("<center><h1 style=\"color:black;\"> Results </h1> </center><hr style=\"height:2px;border-width:0;color:gray;background-color:gray\">\n")

            print("\033[1;32m '-> Creating plots I/O time\033[1;0m") 


    def generate_html_core(self,html_file, f) -> None:
        # create sub HTML files
        print(f"\033[1;34m  '-> Started generating {html_file}\033[1;0m")
        figures_to_html(f, os.path.join(self.path,html_file), self.names)    
        # Mark entry in main.html
        with open(self.filename, "a") as file:
            self.lock.acquire()
            tmp = html_file.replace("_"," ").replace("html"," ").capitalize()
            file.write(f"<h3 ><a href=\"file://{self.path}/{html_file}\" style=\"color:black;\">{tmp}: {len(f)} figures </a>\n")
            self.lock.release()

        print(f" '-> Finished generating {html_file}")


    def generate_html_end(self) -> None:
        # clsoe the file
        with open(self.filename, 'a') as file:
            file.write("<hr style=\"height:2px;border-width:0;color:gray;background-color:gray\">\n")
            if self.names:
                # self.names = list(set(self.names))
                file.write("<br><br> Folders map to:<ul>\n")    
                for i in self.names:
                    file.write(f"<li> Run {self.names.index(i)}: {i} </li>\n")    
                file.write("</ul> \n")
            file.write("</body></html> \n")

        print(f"\nTo see the result call \nopen {self.path}/main.html \n")
        if self.show:
            if platform == "linux" or platform == "linux2":
                if "WSL" in plat.uname().release:
                    os.system(f"powershell.exe start ./{self.outdir}/main.html ")
                else:
                    os.system(f"open {self.path}/main.html \n")

            if "windows" in platform:
                try:
                    os.system(f"powershell.exe start {self.path}/main.html &\n" )
                except:
                    os.system(f"powershell.exe start./{self.path}/main.html")


#**********************************************************************
#*                       1. figures_to_html
#**********************************************************************
def figures_to_html(figs, filename="write_async.html",names=[]) -> None:
    # conf = {  "toImageButtonOptions": {     "format": "svg", "scale":1  }}
    conf = {  "toImageButtonOptions": {     "format": "png", "scale":2 }}
    template = """
    <!DOCTYPE html>
    <html lang="en">
    <head>
    <meta charset="utf-8"/>
    <script>{plotly}</script>
    </head>
    <body>
    {names}
    {plots}
    </body>
    </html>
    """
    html_parts = ""
    for fig in figs:
        if figs.index(fig)+1 == len(figs):
            print("    '-> figure (%i/%i) "%(figs.index(fig)+1,len(figs)), end="\n")
        else:
            print("    '-> \033[1;32mfigure (%i/%i) \033[1;0m"%(figs.index(fig)+1,len(figs)), end="\r")
        html_parts= html_parts + fig.to_html(config=conf,include_plotlyjs=False,include_mathjax = "cdn") + "\n"

    run_names = ""
    if names:                
        run_names ="<br><br> Folders map to:<ul>\n"
        for i in names:
            run_names = run_names + (f"<li> Run {names.index(i)}: {i} </li>\n")    
        run_names = run_names + "</ul> \n"

    template = template.format(plotly=plotly.offline.get_plotlyjs(), plots = html_parts, names = run_names)
    with open(filename, "w") as file:
        file.write(template)





import numpy as np
import glob
from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
from ftio.freq.freq_plot_core import convert_and_plot
from ftio.parse.bandwidth import overlap
from ftio.api.gekkoFs.parse_gekko import parse
import plotly.graph_objects as go
from ftio.freq.helper import format_plot
from ftio.plot.units import set_unit
from ftio.freq.helper import MyConsole

CONSOLE = MyConsole()
CONSOLE.set(True)

def run(files_or_msgs: list, argv:list[str]=["-e", "plotly", "-f", "100"], b_app = [], t_app = []):  # "0.01"] ):
    """Executes ftio on list of files_or_msgs.

    Args:
        files_or_msgs (list): _description_
        argv: command line arguments from ftio
        data_rank
    """

    # parse args
    args = parse_args(argv, "ftio")
    ranks = len(files_or_msgs)

    # Set up data
    data_rank = {
        "avg_thruput_mib": [],
        "end_t_micro": [],
        "start_t_micro": [],
        "hostname": "",
        "pid": 0,
        "io_type":"",
        "req_size": [],
        "total_bytes": 0,
        "total_iops": 0,
        "flush_t": 0, 
    }

    # 1) overlap for rank level metrics
    for file_or_msg in files_or_msgs:
        data_rank, ext = parse(file_or_msg, data_rank, io_type = args.mode[0])

    # 2) exit if no new data
    if not data_rank["avg_thruput_mib"]:
        CONSOLE.print("[red]Terminating prediction ... [/]")
        exit(0)
    
    # 3) Scale if JSON or MsgPack
    scale = [1, 1, 1]
    if "JSON" in ext.upper():
        scale = [1.07 * 1e6, 1e-3, 1e-3]
    elif any(x in ext.upper() for x in ["MSG", "ZMQ"]): 
        scale = [1, 1e-6, 1e-6]
        
    b_rank   = np.array(data_rank["avg_thruput_mib"]) * scale[0]
    t_rank_s = np.array(data_rank["start_t_micro"]) * scale[1]
    t_rank_e = np.array(data_rank["end_t_micro"]) * scale[2]
    if "flush_t" in data_rank:
        data_rank["flush_t"] = data_rank["flush_t"]* scale[2]

    # 4) app level bandwidth
    b, t = overlap(b_rank, t_rank_s, t_rank_e)
        
    # 5) Extend for ZMQ
    if "ZMQ" in ext.upper(): # or use args.zmq
        # extend data
        b_app.extend(b)
        t_app.extend(t)
        t = np.array(t_app[:])
        b = np.array(b_app[:])
    else:
        t = np.array(list(b))
        b = np.array(list(t))

    # 6) plot to check:
    if any(x in args.engine for x in ["mat", "plot"]):
        fig = go.Figure()
        unit, order = set_unit(b)
        fig.add_trace(go.Scatter(x=t, y=b * order, name="App Bandwidth",line={"shape": "hv"}))
        fig.update_layout(xaxis_title="Time (s)", yaxis_title=f"Bandwidth ({unit})")
        fig = format_plot(fig)
        fig.show()

    # 7) set up data
    data = {
        "time": t,
        "bandwidth": b,
        "total_bytes": data_rank["total_bytes"],
        "ranks": ranks,
    }

    # 8) perform prediction
    prediction, dfs = core([data], args)

    # 9) plot and print info
    display_prediction("ftio", prediction)
    convert_and_plot(data, dfs, args)

    return prediction, args, data_rank["flush_t"]


if __name__ == "__main__":
    # absolute path to search all text files_or_msgs inside a specific folder
    # path=r'/d/github/FTIO/examples/API/gekkoFs/JSON/*.json'
    path = r"/d/github/FTIO/examples/API/gekkoFs/MSGPACK/write*.msgpack"
    matched_files_or_msgs = glob.glob(path)
    run(matched_files_or_msgs)



"""Merge Predections

"""
from __future__ import annotations
import time
import numpy as np
from rich.panel import Panel
from ftio.freq.helper import MyConsole

CONSOLE = MyConsole()

def merge_predictions(args, pred_dft:dict, pred_auto:dict)-> dict:
    """Merge prediction if both autocorrelation and freq. technique (dft/wavelet) are executed

    Args:
        args (_type_): _description_
        pred_dft (dict): prediction frequency technique
        pred_auto (dict): prediction autocorrelation

    Returns:
        dict: merge prediction
    """
    CONSOLE.set(args.verbose)
    pred_merged = pred_dft.copy()
    if args.autocorrelation:
        tik = time.time()
        CONSOLE.print(f"[cyan]Merging Started:[/]\n")
        text = f"Merging Autocorrelation and {args.transformation.upper()}\n"
        if pred_auto and  not np.isnan(pred_auto["dominant_freq"]):
            pred_merged["dominant_freq"], pred_merged["conf"], text = merge_core(pred_dft, pred_auto, args.freq, text)
            if pred_merged["dominant_freq"]:
                text += f"Dominant frequency: [blue bold]{np.round(pred_merged['dominant_freq'][-1],4)}[/] Hz -> [blue bold]{np.round(1/pred_merged['dominant_freq'][-1],4)}[/] sec\n"
                text += f"Confidence: [bold]{color_pred(pred_merged['conf'][-1])}{np.round(pred_merged['conf'][-1]*100,2)}[/] %\n"
            pred_auto.pop("candidates", None)
        else:
            text += "[yellow]Autrocorrelation prediction is empty[/]\n"
        CONSOLE.print(Panel.fit(text[:-1], style="white", border_style="blue", title="Merging Predections", title_align='left'))
        CONSOLE.print(f"\n[cyan]Merging finished:[/] {time.time() - tik:.3f} s")

    return pred_merged


def merge_core(pred_dft:dict, pred_auto:dict ,freq:float, text:str) -> tuple[list,list,str]:
    """Merge the predictions

    Args:
        pred_dft (dict): prediction from dft. technique
        pred_auto (dict): prediction from autocorreleation
        freq: sampling frequency
        text (str): display string

    Returns:
        tuple[list,list,str]: dominant frequency, confidence, and display string
    """
    dominant_freq = -1
    conf = 0
    out_freq, out_conf = [],[]
    method = "hits"
    method2 = "cov" # ratio or cov for method hits
    if len(pred_dft["dominant_freq"]) >= 1:
        if "alike" in method:
            alike = (pred_auto["dominant_freq"] - abs(pred_dft["dominant_freq"] - pred_auto["dominant_freq"]))/pred_auto["dominant_freq"]
            text += f"Frequencies [yellow]{np.round(pred_dft['dominant_freq'],4)}[/] Hz match [yellow]{np.round(pred_auto['dominant_freq'],4)}[/] Hz by:\n[yellow]{np.round(alike,4)}[/]%\n\n"
            dominant_index = np.argmax(alike)
            dominant_freq = pred_dft["dominant_freq"][dominant_index]
            conf  = (pred_dft["conf"][dominant_index] + pred_auto["conf"])/2
        elif "hits" in method:
            tol = 2/freq # 2 frequency steps
            hits = np.zeros(len(pred_dft["dominant_freq"]))
            if "ratio" in method:
                alike = np.zeros((len(pred_dft["dominant_freq"]),len(pred_auto["candidates"])))
            else: # cov method
                alike = np.zeros(len(pred_dft["dominant_freq"]))

            # Find perfect hits and caclualte distance of the candiidates to the freq predictions
            for i,f_d in enumerate(pred_dft["dominant_freq"]):
                t_d = 1/f_d
                hits[i] = sum((t_d - tol < pred_auto["candidates"])  & (pred_auto["candidates"] < t_d + tol ))
                if "ratio" in method:
                    alike[i,:] = np.min([pred_auto["candidates"]/t_d, t_d/pred_auto["candidates"]],axis=0) #1 - abs(pred_auto["candidates"] - t_d)/t_d
                else: # cov method
                    alike[i] = 1 - np.abs(np.std(np.append(pred_auto["candidates"],t_d))/np.mean(np.append(pred_auto["candidates"],t_d)))


            # check if there is a predect prediction
            text += f"Perfect hits of [blue]{pred_dft['dominant_freq']}[/] are [blue] {hits}[/]  \n"
            for i,value in enumerate(hits):
                if value == len(pred_auto["candidates"]):
                    text += "[green bold]Perfect Prediction found! [/]\n"
                    dominant_freq = pred_dft["dominant_freq"][i]
                    conf  = 1
                    break

            #No perfect hit, see which better fits
            if "ratio" in method:
                dominant_index = np.argmax(np.sum(alike, axis=1))
            else: # cov method
                dominant_index = np.argmax(alike)
            
            dominant_freq = pred_dft["dominant_freq"][dominant_index]
            # calculate conf:
            agreed_predictions = np.argmax(alike,axis=0)
            if "ratio" in method:
                text += f"Agreed prediction ratio  [blue] {np.sum(alike, axis=1)/len(agreed_predictions)*100}[/] %\n"
                conf = (1 - np.std([pred_dft["conf"][dominant_index],pred_auto["conf"]])/np.mean([pred_dft["conf"][dominant_index],pred_auto["conf"]])
                    + pred_dft["conf"][dominant_index]
                    + pred_auto["conf"]
                    )/3 
            else: 
                text += f"Agreed prediction [blue] {alike*100}[/] %\n"
                conf = (alike[dominant_index] + pred_dft["conf"][dominant_index] + pred_auto["conf"])/3 
                # text += f"[red bold]{alike[dominant_index]} + {pred_dft['conf'][dominant_index]} + {pred_auto['conf']}[/]\n"
                
            if conf >= 0.2:
                out_freq, out_conf = [dominant_freq], [conf]
            else:
                out_freq, out_conf = [], []
                text += "[red bold]Too low confidence, no dominant freq[/]\n"
                
    else:# no dft result
        dominant_freq = pred_auto["dominant_freq"]
        conf = pred_auto["conf"]/3
        text += "Confidence: [red] Warning! Low confidence! [/]\n"
        out_freq, out_conf = [dominant_freq], [conf]

    return out_freq, out_conf, text


def color_pred(conf:float)-> str:
    """highlight color according to value

    Args:
        conf (float): value between [0,1]

    Returns:
        string: color to be used with Console (from rich)
    """
    if conf >= 0.8:
        color = "[green]"
    elif conf >= 0.6:
        color = "[blue]"
    elif conf >= 0.3:
        color = "[yellow]"
    else:
        color = "[red]"
    return color



import dash
import numpy as np
import plotly.graph_objects as go
from dash import MATCH, Input, Output, State, dcc, html
from dash_extensions.enrich import DashProxy, Serverside
from plotly_resampler import FigureResampler
from plotly_resampler.aggregation import MinMaxAggregator, MinMaxOverlapAggregator, NoGapHandler

# -------
# TODO: Check if this problem is still not fixed or feature is now implemented
# Solves a problem for a still open issue
# Issue: line plot ends cutoff after last visible point #257
# https://github.com/predict-idlab/plotly-resampler/issues/257
from plotly_resampler.aggregation.plotly_aggregator_parser import PlotlyAggregatorParser
from trace_updater import TraceUpdater

import ftio.plot.dash_files.constants.id as id
import ftio.plot.dash_files.constants.io_mode as io_mode
import ftio.plot.dash_files.constants.legend_group as legend_group
from ftio.plot.dash_files.data_source import DataSource, FileData


class patched_parser(PlotlyAggregatorParser):
    @staticmethod
    def get_start_end_indices(hf_trace_data, axis_type, start, end):
        start_idx, end_idx = PlotlyAggregatorParser.get_start_end_indices(
            hf_trace_data, axis_type, start, end
        )
        start_idx = min(max(0, start_idx - 1), max(0, start_idx - 2))
        length_x = len(hf_trace_data["x"])
        end_idx = max(min(length_x, end_idx + 1), min(length_x, end_idx + 2))
        return start_idx, end_idx


from plotly_resampler.figure_resampler import figure_resampler_interface

figure_resampler_interface.PlotlyAggregatorParser = patched_parser
# end of problem solving issue #257
# -------


def _create_id_figure(data: DataSource, filename: str = ""):
    return f"{data.io_mode}-{filename}"


def _find_x_min_and_max(file_data: FileData, data: DataSource) -> tuple[float, float]:
    x_min = np.inf
    x_max = -np.inf
    if file_data.data_actual_is_not_empty:
        x_min = min(x_min, min(file_data.actual_time_overlap, default=0.0))
        x_max = max(x_max, max(file_data.actual_time_overlap, default=0.0))
        if data.individual_is_selected:
            x_min = min(x_min, min(file_data.actual_time_overlap_individual, default=0.0))
            x_max = max(x_max, max(file_data.actual_time_overlap_individual, default=0.0))
    if file_data.data_required_is_not_empty:
        x_min = min(x_min, min(file_data.required_time_overlap, default=0.0))
        x_max = max(x_max, max(file_data.required_time_overlap, default=0.0))
        if data.individual_is_selected:
            x_min = min(x_min, min(file_data.required_time_overlap_individual, default=0.0))
            x_max = max(x_max, max(file_data.required_time_overlap_individual, default=0.0))
    return (x_min, x_max)


def _add_trace_average(fig: FigureResampler, file_data: FileData) -> None:
    if file_data.data_actual_is_not_empty:
        fig.add_trace(
            go.Scatter(
                mode="lines",
                line={"shape": "hv"},
                fill="tozeroy",
                legendgroup=legend_group.ACTUAL_AVERAGE,
                legendgrouptitle_text=legend_group.ACTUAL_AVERAGE_TITLE,
                name=file_data.name,
            ),
            hf_x=file_data.actual_time_overlap,
            hf_y=file_data.actual_bandwidth_overlap_average,
        )
    if file_data.data_required_is_not_empty:
        fig.add_trace(
            go.Scatter(
                mode="lines",
                line={"shape": "hv"},
                fill="tozeroy",
                legendgroup=legend_group.REQUIRED_AVERAGE,
                legendgrouptitle_text=legend_group.REQUIRED_AVERAGE_TITLE,
                name=file_data.name,
            ),
            hf_x=file_data.required_time_overlap,
            hf_y=file_data.required_bandwidth_overlap_average,
        )


def _add_trace_sum(fig: FigureResampler, file_data: FileData) -> None:
    if file_data.data_actual_is_not_empty:
        fig.add_trace(
            go.Scatter(
                mode="lines",
                line={"shape": "hv"},
                fill="tozeroy",
                legendgroup=legend_group.ACTUAL_SUM,
                legendgrouptitle_text=legend_group.ACTUAL_SUM_TITLE,
                name=file_data.name,
            ),
            hf_x=file_data.actual_time_overlap,
            hf_y=file_data.actual_bandwidth_overlap_sum,
        )
    if file_data.data_required_is_not_empty:
        fig.add_trace(
            go.Scatter(
                mode="lines",
                line={"shape": "hv"},
                fill="tozeroy",
                legendgroup=legend_group.REQUIRED_SUM,
                legendgrouptitle_text=legend_group.REQUIRED_SUM_TITLE,
                name=file_data.name,
            ),
            hf_x=file_data.required_time_overlap,
            hf_y=file_data.required_bandwidth_overlap_sum,
        )


def _add_trace_indiviual(fig: FigureResampler, file_data: FileData) -> None:
    if file_data.data_actual_is_not_empty:
        fig.add_trace(
            go.Scatter(
                mode="lines",
                line={"shape": "hv"},
                fill="tozeroy",
                legendgroup=legend_group.ACTUAL_INDIVIDUAL,
                legendgrouptitle_text=legend_group.ACTUAL_INDIVIDUAL_TITLE,
                name=file_data.name,
            ),
            hf_x=file_data.actual_time_overlap_individual,
            hf_y=file_data.actual_bandwidth_overlap_individual,
        )
    if file_data.data_required_is_not_empty:
        fig.add_trace(
            go.Scatter(
                mode="lines",
                line={"shape": "hv"},
                fill="tozeroy",
                legendgroup=legend_group.REQUIRED_INDIVIDUAL,
                legendgrouptitle_text=legend_group.REQUIRED_INDIVIDUAL_TITLE,
                name=file_data.name,
            ),
            hf_x=file_data.required_time_overlap_individual,
            hf_y=file_data.required_bandwidth_overlap_individual,
        )


def _add_trace_invisible_for_complete_presentation(
    fig: FigureResampler, file_data: FileData, data: DataSource
) -> None:
    x_min, x_max = _find_x_min_and_max(file_data, data)
    fig.add_trace(
        go.Scatter(
            showlegend=False,
            hoverinfo="skip",
            opacity=0.0,
        ),
        hf_x=[x_min, x_max],
        hf_y=[0.0, 0.0],
    )


def _add_traces(fig: FigureResampler, file_data: FileData, data: DataSource) -> None:
    _add_trace_average(fig, file_data)
    _add_trace_sum(fig, file_data)
    if data.individual_is_selected:
        _add_trace_indiviual(fig, file_data)
    _add_trace_invisible_for_complete_presentation(fig, file_data, data)


def _update_layout(fig: FigureResampler, file_data: FileData, data: DataSource) -> None:
    if data.merge_plots_is_selected:
        title = "{} collected".format(io_mode.MODE_STRING_BY_MODE[data.io_mode])
    else:
        title = "{} Ranks (Run {}: {})".format(file_data.rank, file_data.run, file_data.name)

    fig.update_layout(
        margin=dict(l=10, r=10, t=50, b=70),
        yaxis_rangemode="nonnegative",
        barmode="stack",
        xaxis_title="Time (s)",
        yaxis_title="Transfer Rate (MB/s)",
        font=dict(family=data.fontfamily, size=data.fontsize),
        width=data.width_figure,
        height=data.height_figure,
        title=title,
    )


def _update_axes(fig: FigureResampler) -> None:
    fig.update_xaxes(
        ticks="outside",
        tickcolor="black",
        ticklen=6,
        minor=dict(
            ticklen=3,
            tickcolor="black",
            tickmode="auto",
            nticks=10,
            showgrid=True,
        ),
    )
    fig.update_yaxes(
        ticks="outside",
        tickcolor="black",
        ticklen=6,
        minor=dict(
            ticklen=3,
            tickcolor="black",
            tickmode="auto",
            nticks=10,
            showgrid=True,
        ),
    )


def _create_separate_figures(filenames: list[str], data: DataSource) -> dict:
    figure_by_id_figure = dict()

    for filename in filenames:
        file_data = data.file_data_by_filename[filename]
        fig = FigureResampler(
            default_n_shown_samples=data.n_shown_samples,
            default_downsampler=MinMaxAggregator(),
            default_gap_handler=NoGapHandler(),
        )
        _add_traces(fig, file_data, data)

        _update_layout(fig, file_data, data)
        _update_axes(fig)

        id_figure = _create_id_figure(data, filename)
        figure_by_id_figure[id_figure] = fig
    return figure_by_id_figure


def _create_one_common_figure(filenames: list[str], data: DataSource) -> dict:
    figure_by_id_figure = dict()

    fig = FigureResampler(
        default_n_shown_samples=data.n_shown_samples,
        default_downsampler=MinMaxAggregator(),
        default_gap_handler=NoGapHandler(),
    )

    for filename in filenames:
        file_data = data.file_data_by_filename[filename]
        _add_traces(fig, file_data, data)

    _update_layout(fig, file_data, data)
    _update_axes(fig)
    id_figure = _create_id_figure(data)
    figure_by_id_figure[id_figure] = fig
    return figure_by_id_figure


def _append_merged_plot(
    div_children: list[html.Div], figure_by_id_figure: dict[str, FigureResampler], data: DataSource
) -> html.Div:
    id_figure = _create_id_figure(data)
    new_child = html.Div(
        children=[
            dcc.Graph(
                id={"type": id.TYPE_DYNAMIC_GRAPH, "index": id_figure},
                figure=figure_by_id_figure[id_figure],
                mathjax=True, responsive=True,
            ),
            TraceUpdater(
                id={
                    "type": id.TYPE_DYNAMIC_UPDATER_BY_IO_MODE[data.io_mode],
                    "index": id_figure,
                },
                gdID=f"{id_figure}",
            ),
        ],
    )
    div_children.append(new_child)
    return div_children


def _append_each_figure_separately(
    div_children: list[html.Div],
    figure_by_id_figure: dict[str, FigureResampler],
    filenames: list[str],
    data: DataSource,
) -> html.Div:
    for rank in data.ranks_unique:
        for filename in filenames:
            if data.file_data_by_filename[filename].rank != rank:
                continue
            id_figure = _create_id_figure(data, filename)
            new_child = html.Div(
                children=[
                    dcc.Graph(
                        id={"type": id.TYPE_DYNAMIC_GRAPH, "index": id_figure},
                        figure=figure_by_id_figure[id_figure],
                        mathjax=True, responsive=True,
                    ),
                    TraceUpdater(
                        id={
                            "type": id.TYPE_DYNAMIC_UPDATER_BY_IO_MODE[data.io_mode],
                            "index": id_figure,
                        },
                        gdID=f"{id_figure}",
                    ),
                ],
            )
            div_children.append(new_child)
    return div_children


def get_io_mode_specific_callbacks(app: DashProxy, data: DataSource) -> None:
    """Defines the callbacks to a specific io_mode which is indirectly given by the data.

    Args:
        app (DashProxy): The dash app
        data (DataSource): data_source belonging to the io_mode
    """

    @app.callback(
        Output(id.STORE_FIGURES_BY_IO_MODE[data.io_mode], "data"),
        Input(id.DROPDOWN_FILE, "options"),
    )
    def create_figures(filenames: list[str]) -> dict[str, FigureResampler]:
        figure_by_id_figure = dict()

        if data.merge_plots_is_selected:
            figure_by_id_figure = _create_one_common_figure(filenames, data)
        else:
            figure_by_id_figure = _create_separate_figures(filenames, data)

        return Serverside(figure_by_id_figure)

    @app.callback(
        Output(id.DIV_IO_BY_IO_MODE[data.io_mode], "children"),
        State(id.CHECKLIST_IO_MODE, "value"),
        State(id.DROPDOWN_FILE, "value"),
        State(id.STORE_FIGURES_BY_IO_MODE[data.io_mode], "data"),
        Input(id.BUTTON_SHOW, "n_clicks"),
        prevent_initial_call=True,
    )
    def fill_div_graph(
        io_modes: list[str],
        filenames: list[str],
        figure_by_id_figure: dict[str, FigureResampler],
        n_clicks: int,
    ):
        if data.io_mode not in io_modes:
            return []

        div_children = [
            html.Div(
                dcc.Markdown(
                    "### {}".format(io_mode.MODE_STRING_BY_MODE[data.io_mode]),
                    mathjax=True,
                )
            )
        ]

        if data.merge_plots_is_selected:
            div_children = _append_merged_plot(div_children, figure_by_id_figure, data)
        else:
            div_children = _append_each_figure_separately(
                div_children, figure_by_id_figure, filenames, data
            )

        return div_children

    @app.callback(
        Output(
            {"type": id.TYPE_DYNAMIC_UPDATER_BY_IO_MODE[data.io_mode], "index": MATCH},
            "updateData",
        ),
        Input({"type": id.TYPE_DYNAMIC_GRAPH, "index": MATCH}, "relayoutData"),
        State({"type": id.TYPE_DYNAMIC_GRAPH, "index": MATCH}, "id"),
        State(id.STORE_FIGURES_BY_IO_MODE[data.io_mode], "data"),
        prevent_initial_call=True,
        memoize=True,
    )
    def update_figure(
        relayoutdata: dict,
        id: dict[str, str],
        figure_by_id_figure: dict[str, FigureResampler],
    ):
        fig = figure_by_id_figure[id["index"]]
        if fig is None:
            return dash.no_update
        return fig.construct_update_data(relayoutdata)



import numpy as np
from ftio.parse.overlap_thread import overlap_thread
from numba import jit


class Bandwidth:
    """this class contains the bandwidth + time metrics at the
        different levels:
        1) request level: subscripted ind (for individual requests)
    _ovr
        2) Rank level metrics: avrerage (avr) or sum. Represent either the
        sum/avr of the ind level metrics during a phase.

        3) Application level metrics:  subscripted overlap. Calculated by
        overlapping the previous metrics. The most accurate one is ind, however,
        it has the highest overhead.
    """

    def __init__(self, b, io_type, args):  # b = values["bandwidth"]
        # App level
        self.b_overlap_sum = []
        self.b_overlap_avr = []
        self.t_overlap = []
        self.b_overlap_ind = []
        self.t_overlap_ind = []
        # Rank level
        self.b_rank_sum = []
        self.b_rank_avr = []
        self.t_rank_s = []
        self.t_rank_e = []
        # Ind level
        self.b_ind = []
        self.t_ind_s = []
        self.t_ind_e = []

        # Application level metrics
        # if "n_overlap" in b:
        # 	self.n_overlap.extend(b["n_overlap"])
        if "b_overlap_sum" in b:
            self.b_overlap_sum.extend(b["b_overlap_sum"])
        if "b_overlap_avr" in b:
            self.b_overlap_avr.extend(b["b_overlap_avr"])
        if "t_overlap" in b:
            self.t_overlap.extend(b["t_overlap"])
        else:
            if "b_overlap_sum" in b:
                self.t_overlap.extend(np.zeros(len(b["b_overlap_sum"])))
            elif "b_overlap_avr" in b:
                self.t_overlap.extend(np.zeros(len(b["b_overlap_avr"])))

        # Rank level metrics
        # remove later
        # ---------------------------------
        if "b" in b and "async" in io_type:
            print("Compatibility mode")
            self.b_rank_sum.extend(b["b"])
            self.b_rank_avr.extend(np.zeros(len(b["b"])))
            self.t_rank_s.extend(np.zeros(len(b["b"])))
            self.t_rank_e.extend(np.zeros(len(b["b"])))
        elif "b" in b and "sync" in io_type:
            print("Compatibility mode")
            self.b_rank_avr.extend(b["b"])
            self.b_rank_sum.extend(np.zeros(len(b["b"])))
            self.t_rank_s.extend(np.zeros(len(b["b"])))
            self.t_rank_e.extend(np.zeros(len(b["b"])))
        # ---------------------------------

        if args.avr or args.sum:
            # 1) assign rank level metric
            if args.sum:
                if "b_rank_sum" in b:
                    self.b_rank_sum.extend(b["b_rank_sum"])

            if args.avr:
                if "b_rank_avr" in b:
                    self.b_rank_avr.extend(b["b_rank_avr"])

            if "t_rank_s" in b:
                self.t_rank_s.extend(b["t_rank_s"])
                if "t_rank_e" in b and b["t_rank_e"]:
                    self.t_rank_e.extend(b["t_rank_e"])
                else:
                    if b["t_rank_s"]:
                        self.t_rank_e.extend(b["t_rank_s"][1:])
                        self.t_rank_e.append(b["t_rank_s"][-1])
                        b["t_rank_e"] = self.t_rank_e
                    else:
                        pass
            else:
                if "b_rank_sum" in b:
                    self.t_rank_s.extend(np.zeros(len(b["b_rank_sum"])))
                    self.t_rank_e.extend(np.zeros(len(b["b_rank_sum"])))
                elif "b_rank_avr" in b:
                    self.t_rank_s.extend(np.zeros(len(b["b_rank_avr"])))
                    self.t_rank_e.extend(np.zeros(len(b["b_rank_avr"])))

            # 2) Calculate bandwidth overlapping at rank level
            if "t_rank_s" in b  and "b_rank_avr" in b and "b_overlap_avr" not in b and args.avr:
                self.b_overlap_avr, self.t_overlap = overlap(
                    b["b_rank_avr"], b["t_rank_s"], b["t_rank_e"]
                )
            if "t_rank_s" in b and "b_rank_sum" in b and "b_overlap_sum" not in b and args.sum:
                self.b_overlap_sum, self.t_overlap = overlap(
                    b["b_rank_sum"], b["t_rank_s"], b["t_rank_e"]
                )

        # overlapping thread level
        if args.ind:
            # Thread level metrics
            if "b_ind" in b:
                self.b_ind.extend(b["b_ind"])
                #! overlap ind
                self.b_overlap_ind, self.t_overlap_ind = overlap(
                    b["b_ind"], b["t_ind_s"], b["t_ind_e"]
                )

            if "t_ind_s" in b:
                self.t_ind_s.extend(b["t_ind_s"])

            if "t_ind_e" in b:
                self.t_ind_e.extend(b["t_ind_e"])

        self.app_ind = max(self.b_overlap_ind) if self.b_overlap_ind else -1
        self.app_avr = max(self.b_overlap_avr) if self.b_overlap_avr else -1
        self.app_sum = max(self.b_overlap_sum) if self.b_overlap_sum else -1

        # statistics:
        self.weighted_harmonic_mean = self.assign(b, "weighted_harmonic_mean")
        self.harmonic_mean = self.assign(b, "harmonic_mean")
        self.arithmetic_mean = self.assign(b, "arithmetic_mean")
        self.median = self.assign(b, "median")
        self.max = self.assign(b, "max")
        self.min = self.assign(b, "min")

        self.weighted_avr_harmonic_mean = self.assign(b, "weighted_avr_harmonic_mean")
        self.harmonic_avr_mean = self.assign(b, "harmonic_avr_mean")
        self.arithmetic_avr_mean = self.assign(b, "arithmetic_avr_mean")
        self.median_avr = self.assign(b, "median_avr")
        self.max_avr = self.assign(b, "max_avr")
        self.min_avr = self.assign(b, "min_avr")

        self.weighted_sum_harmonic_mean = self.assign(b, "weighted_sum_harmonic_mean")
        self.harmonic_sum_mean = self.assign(b, "harmonic_sum_mean")
        self.arithmetic_sum_mean = self.assign(b, "arithmetic_sum_mean")
        self.median_sum = self.assign(b, "median_sum")
        self.max_sum = self.assign(b, "max_sum")
        self.min_sum = self.assign(b, "min_sum")

    def assign(self, b, name):
        if name in b:
            return b[name] if not np.isnan(b[name]) else -1
        else:
            return -1


#! ----------------------- I/O analysis ------------------------------
# **********************************************************************
# *                       1. Overlap
# **********************************************************************


def overlap(b, t_s, t_e):
    t_s = np.array(t_s)
    t_e = np.array(t_e)
    b = np.array(b)
    id_s = np.argsort(t_s)
    id_e = np.argsort(t_e)
    try:
        b_overlap, t_overlap = overlap_core(b, t_s, t_e, id_s, id_e)
    except:
        b_overlap, t_overlap = overlap_core_safe(b, t_s, t_e, id_s, id_e)

    return list(b_overlap), list(t_overlap)


@jit(nopython=True, cache=True)
def overlap_core(b, t_s, t_e, id_s, id_e):
    agg_phases = len(t_s)

    b_out = np.zeros(2 * agg_phases)
    t_out = np.zeros(2 * agg_phases)

    b_tmp = 0
    k_s = 0
    k_e = 0
    counter = 0

    while k_s < agg_phases or k_e < agg_phases:
        if k_s == agg_phases or t_e[id_e[k_e]] < t_s[id_s[k_s]]:
            b_tmp = b_tmp - b[id_e[k_e]]
            # t_out.append(t_e[id_e[k_e]])
            t_out[counter] = t_e[id_e[k_e]]
            k_e += 1
        else:
            b_tmp = b_tmp + b[id_s[k_s]]
            # t_out.append(t_s[id_s[k_s]])
            t_out[counter] = t_s[id_s[k_s]]
            k_s += 1
        # b_out.append(b_tmp)
        b_out[counter] = b_tmp
        counter += 1

    return b_out, t_out

def overlap_core_safe(b, t_s, t_e, id_s, id_e):
    agg_phases = len(t_s)

    b_out = []
    t_out = []
    b_tmp = 0
    k_s = 0
    k_e = 0

    while k_s < agg_phases or k_e < agg_phases:
        if k_s == agg_phases or t_e[id_e[k_e]] < t_s[id_s[k_s]]:
            b_tmp = b_tmp - b[id_e[k_e]]
            t_out.append(t_e[id_e[k_e]])
            k_e += 1
        else:
            b_tmp = b_tmp + b[id_s[k_s]]
            t_out.append(t_s[id_s[k_s]])
            k_s += 1
        b_out.append(b_tmp)

    return b_out, t_out



""" Extracts time behavior form parsed data

"""
import pandas as pd 


def get_time_behavior(df) -> list[dict]:
    """Get the time behavior

    Args:
        df (dataframe): obtained from scales.py
    """
    out = []
    files = [int(i) for i in pd.unique(df[0]["number_of_ranks"])]
    for i in files:
        ranks = df[1]["number_of_ranks"].isin([i])
        if len(df[1]["file_index"][ranks]) != 0:
            for j in range(0, int(df[1]["file_index"][ranks].max() + 1)):
                # print(f"  \033[1;32mRanks {i}\033[1;0m")
                file_index = df[1]["file_index"][ranks].isin([j])
                time = df[1]["t_overlap"][ranks][file_index].to_numpy()
                bandwidth = df[1]["b_overlap_avr"][ranks][file_index].to_numpy()
                try:
                    total_bytes = df[0]["total_bytes"].to_numpy()
                    total_bytes = int(float(total_bytes[-1]))
                except ValueError:
                    total_bytes = 0
                    # expe.center()np.sum(bandwidth * (np.concatenate([time[1:], time[-1:]]) - time)
                tmp = {
                    "time": time,
                    "bandwidth": bandwidth,
                    "total_bytes": total_bytes,
                    "ranks": i
                    }
                out.append(tmp)
    return out



"""Performs prediction with Pools (ProcessPoolExecutor) and a callback mechanism"""
from __future__ import annotations
import ftio.prediction.monitor as pm
from ftio.prediction.probability_analysis import find_probability
from ftio.prediction.helper import print_data, export_extrap
from ftio.prediction.analysis import ftio_process
from ftio.prediction.async_process import handle_in_process

# from ftio.prediction.async_process import handle_in_process


def predictor_with_processes(
    filename, data, queue, count, hits, start_time, aggregated_bytes, args
):
    """performs prediction in ProcessPoolExecuter. FTIO is a submitted future and probability is calculated as a callback

    Args:   
        filename (str): name of file
        data (Manager().list): List of dicts with all predictions so far
        queue (Manager().Queue): queue for FTIO data
        count (Manager().Value): number of prediction
        hits (Manager().Value): hits indicating how often a dominant frequency was found
        start_time (Manager().Value): start time window for ftio
        aggregated_bytes (Manager().Value): total bytes transferred so far
        args (list[str]): additional arguments passed to ftio
    """
    procs = []
    # Init: Monitor a file
    stamp,_ = pm.monitor(filename, "")

    # Loop and predict if changes occur
    try:
        while True:
            # monitor
            stamp, procs = pm.monitor(filename, stamp, procs)
            # launch prediction_process
            procs.append(
                handle_in_process(
                    prediction_process,
                    args=(data, queue, count, hits, start_time, aggregated_bytes, args)
                )
            )
    except KeyboardInterrupt:
        print_data(data)
        export_extrap(data)
        print("-- done -- ")


def prediction_process(
    data, queue, count, hits, start_time, aggregated_bytes, args: list[str], msgs=None) -> None:
    """Performs prediction made up of two part: (1) Executes FTIO and (2) appends to data the value

    Args:
        data (Manager().list): List of dicts with all predictions so far
        queue (Manager().Queue): queue for FTIO data
        count (Manager().Value): number of prediction
        hits (Manager().Value): hits indicating how often a dominant frequency was found
        start_time (Manager().Value): start time window for ftio
        aggregated_bytes (Manager().Value): total bytes transferred so far
        args (list[str]): additional arguments passed to ftio.py
    """
    ftio_process(queue, count, hits, start_time, aggregated_bytes, args, msgs)
    while not queue.empty():
        data.append(queue.get())

    _ = find_probability(data)



from ftio.parse.simrun import Simrun
from ftio.parse.msgpack_reader import extract

class ParseMsgpack:
    def __init__(self, path):
        self.path = path
        if self.path[-1] == "/":
            self.path = self.path[:-1]

    def to_simrun(self, args, index = 0):
        """Convert to Simrun class
        Args:
            ars (argparse): command line arguments
            index: file index in case several files are passed
        Returns:
            Simrun: Simrun object
        """
        file = self.path
        data = extract(file)

        return Simrun(data,'msgpack', file, args, index)


# This file is part of the FTIO software 
#
# Copyright (c) 2023-2024, Technical University of Darmstadt, Germany
#
# This software may be modified and distributed under the terms of a BSD-style license.
# See the LICENSE file in the base directory for details.

__title__ = "FTIO"
__version__ = "0.0.4"
__license__ = "BSD 3"
__description__ = "FTIO: Capturing Periodic I/O Using Frequency Techniques"
__copyright__ = "Copyright (c) 2023-2024 Technical University of Darmstadt, Darmstadt, Germany"
__documentation_link__ = f"https://github.com/tuda-parallel/FTIO/tree/v{__version__}/docs"
__current_version_api__ = "https://pypi.org/pypi/extrap/json"
__developed_by_html__ = 'FTIO is developed by <a href="https://www.parallel.informatik.tu-darmstadt.de/">' 
__support_email__ = "ahmad.tarraf@tu-darmstadt.de"
__repo__='https://github.com/tuda-parallel/FTIO'



from ftio.parse.simrun import Simrun
import jsonlines

class ParseJsonl:
    def __init__(self, path):
        self.path = path
        if self.path[-1] == "/":
            self.path = self.path[:-1]

    def to_simrun(self, args, index = 0):
        """Convert to Simrun class
        Args:
            ars (argparse): command line arguments
            index: file index in case several files are passed
        Returns:
            Simrun: Simrun object
        """
        file = self.path
        with jsonlines.open(file, "r") as jsonl_f:
            data = [obj for obj in jsonl_f]

        return Simrun(data, "jsonl", file, args, index)



"""Contains DFT methods and accuracy calculation 
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from ftio.prediction.unify_predictions import  color_pred
from ftio.prediction.helper import get_dominant_and_conf
from ftio.freq.helper import MyConsole


CONSOLE = MyConsole()


#!################
#! DFT flavors
#!################
# Wrapper
def dft(b):
    return numpy_dft(b)


#1) Custome implementation
def dft_fast(b):
    N = len(b)
    X = np.repeat(complex(0, 0), N)  # np.zeros(N)
    for k in range(0, N):
        for n in range(0, N):
            X[k] = X[k] + b[n] * np.exp((-2 * np.pi * n * k / N) * 1j)
    return X

#2) numpy DFT
def numpy_dft(b):
    return np.fft.fft(b)

#3) DFT with complex
def dft_slow(b):
    N = len(b)
    n = np.arange(N)
    k = n.reshape((N, 1))
    e = np.exp(-2j * np.pi * k * n / N)
    X = np.dot(e, b)
    return X


#!################
#! DFT Precision
#!################
def precision_dft(
    amp: np.ndarray, phi:np.ndarray, dominant_index:np.ndarray, b_sampled:np.ndarray, t_disc:np.ndarray, freq_arr:np.ndarray, plt_engine:str
) -> str:
    """calculates the precision of the dft

    Args:
        amp (np.ndarray): amplitude array from DFT
        phi (np.ndarray): phase array from DFT
        dominant_index (np.ndarray): index/indecies of dominant frequency/frequencies
        b_sampled (np.ndarray): discretized bandwdith
        t_disc (np.ndarray): discretized time (constant step size). Start at t_0
        freq_arr (np.ndarray): frequency array 
        plt_engine (str): comand line specific plot engine

    Returns:
        str: precision 
    """
    # 
    showplot = False
    text = ""
    if showplot and ("mat" in plt_engine or "plotly" in plt_engine):
        plt.figure(figsize=(10, 5))
    dc_offset = np.zeros(len(amp))
    for index in dominant_index:
        x = dc_offset + 2 * (1 / len(amp)) * amp[index] * np.cos(
            2 * np.pi * np.arange(0, len(amp)) * (index) / (len(amp)) + phi[index]
        )
        x[x < 0] = 0
        x_2 = x.copy()
        total = np.sum(b_sampled)
        for i, _ in enumerate(x):
            if x[i] > b_sampled[i]:
                x_2[i] = b_sampled[i]
                x[i] = -x[i] + b_sampled[i]

        text += f"Precision of [cyan]{freq_arr[index]:.2f}[/] Hz is [cyan]{float(np.sum(x)) / total * 100:.2f}% [/]"
        text += f"(Positive only: [cyan]{float(np.sum(x_2)) / total * 100:.2f}%[/])\n"

        if showplot and ("mat" in plt_engine or "plotly" in plt_engine):
            plt.plot(t_disc, x, label=f"f = {freq_arr[index]:.2f} Hz")

    if showplot and ("mat" in plt_engine or "plotly" in plt_engine):
        plt.plot(t_disc, b_sampled, label="b_sampled")
        plt.ticklabel_format(axis="y", style="sci", scilimits=(-5, 3))
        plt.ticklabel_format(axis="x", style="sci", scilimits=(-5, 3))
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)
        plt.ylabel("Amplitude (B/s)", fontsize=17)
        plt.xlabel("Time (s)", fontsize=17)
        plt.grid(True)
        plt.legend(loc="upper left", ncol=2, fontsize=13)
        plt.tight_layout()
        plt.show()

    return text


def prepare_plot_dfs(
    k,
    freq,
    freq_arr,
    conf,
    dominant_index,
    amp,
    phi,
    b_sampled,
    time_b,
    ranks,
    bandwidth,
) -> tuple[list, list, list, list]:
    df0 = []
    df1 = []
    df2 = []
    df3 = []
    N = len(b_sampled)

    df3.append(
        pd.DataFrame(
            {
                "dominant": freq_arr[dominant_index],
                "k": dominant_index,
                "conf": conf[dominant_index],
                "ranks": np.repeat(ranks, len(dominant_index)),
            }
        )
    )
    df0.append(
        pd.DataFrame(
            {
                "A": amp,
                "phi": phi,
                "b_sampled": b_sampled,
                "ranks": np.repeat(ranks, N),
                "k": np.arange(0, N),
                "freq": freq_arr,
                "T": np.concatenate([np.array([0]), 1 / freq_arr[1:]]),
                "conf": conf,
            }
        )
    )
    df1.append(
        pd.DataFrame(
            {
                "t_start": time_b[0],
                "t_end": time_b[-1],
                "T_s": 1 / freq,
                "N": N,
                "ranks": ranks,
            },
            index=[k],
        )
    )
    df2.append(
        pd.DataFrame(
            {"b": bandwidth, "t": time_b, "ranks": np.repeat(ranks, len(time_b))}
        )
    )
    return df0, df1, df2, df3


def display_prediction(argv: list[str]|str, prediction: dict) -> None:
    """Displays the result of the prediction from ftio

    Args:
        argv (list[str]): command line arguments
        prediction (dict): the result from ftio
    """

    if isinstance(argv,list):
        func_name = argv[0][argv[0].rfind("/") + 1:]
    else:
        func_name = argv

    if "ftio" in func_name:
        if prediction:
            freq, conf = get_dominant_and_conf(prediction)
            if not np.isnan(freq):
                CONSOLE.info(
                    f"[cyan underline]Prediction results:[/]\n[cyan]Frequency:[/] {freq:.3e} Hz"
                    f"[cyan] ->[/] {np.round(1/freq,4)} s\n"
                    f"[cyan]Confidence:[/] {color_pred(conf)}"
                    f"{np.round(conf*100,2)}[/] %\n"
                )
            else:
                CONSOLE.info(
                        "[cyan underline]Prediction results:[/]\n"
                        "[red]No dominant frequency found[/]\n"
                    )


"""Parse text file contaiting three fields: 
the bandwidth, the start time, and the end time.
"""
from ftio.parse.simrun import Simrun
from ftio.parse.txt_reader import extract

class ParseCustom:
    """class to parse txt file
    """
    def __init__(self, path):
        self.path = path
        if self.path[-1] == "/":
            self.path = self.path[:-1]

    def to_simrun(self, args, index:int = 0):
        """Convert to Simrun class
        Args:
            ars (argparse): command line arguments
            index: file index in case several files are passed
        Returns:
            Simrun: Simrun object
        """
        dataframe, ranks = extract(self.path, args, True)

        return Simrun(dataframe,'txt',str(ranks), args, index)





"""
# Time Window functions
This file contains function that allow modifying and setting the data according the the time window:
- data_in_time_window: cuts the data according the start and end time specified by the arguments.
"""

import numpy as np


def data_in_time_window(
    args, bandwidth: np.ndarray, time_b: np.ndarray, total_bytes: int, ranks: int = 0
) -> tuple[np.ndarray, np.ndarray, str]:
    """Cuts the data according the start and end time specified by the arguments.

    Args:
        args (_type_): argparse
        bandwidth (np.ndarray)
        time_b (np.ndarray)
        total_bytes (int)
        ranks (int, optional). Defaults to 0.

    Returns:
        tuple[np.ndarray,np.ndarray,str]: cut bandwidth and time + text
    """
    text = f"Ranks: [cyan]{ranks}[/]\n"
    ignored_bytes = total_bytes
    # shorten data according to start time
    if args.ts:
        indices = np.where(time_b >= args.ts)
        time_b = time_b[indices]
        bandwidth = bandwidth[indices]
        total_bytes = int(
            np.sum(bandwidth * (np.concatenate([time_b[1:], time_b[-1:]]) - time_b))
        )
        text += f"[green]Start time set to {args.ts:.2f}[/] s\n"
    else:
        text += f"Start time: [cyan]{time_b[0]:.2f}[/] s \n"

    # shorten data according to end time
    if args.te:
        indices = np.where(time_b <= args.te)
        time_b = time_b[indices]
        bandwidth = bandwidth[indices]
        total_bytes = int(
            np.sum(bandwidth * (np.concatenate([time_b[1:], time_b[-1:]]) - time_b))
        )
        text += f"[green]End time set to {args.te:.2f}[/] s\n"
    else:
        text += f"End time: [cyan]{time_b[-1]:.2f}[/] s\n"

    # ignored bytes
    ignored_bytes = ignored_bytes - total_bytes
    text += f"Total bytes: [cyan]{total_bytes:.2e} bytes[/]\n"
    text += f"Ignored bytes: [cyan]{ignored_bytes:.2e} bytes[/]\n"

    return bandwidth, time_b, text



"""
Parses Darshan object to simrun
This function can be also executed as a standalone. Just call:
> python3 darshan_reader.py FILE

Returns:
    list[dict]: _description_
"""
import sys
import time
import darshan
from rich.console import Console
import pandas as pd


def extract(path, args) -> tuple[dict, int]:
    """extracts Darshan file and generates dictionary with relevant keys

    Args:
        path (str): filename
        args (Argparse): optional arguments

    Returns:
        tuple[dict,int]:
            1. dictionary with relevant files
            2. number of ranks
    """
    dataframe, ranks, time_total = extract_data(path, args)
    write, read, time_io = extract_darshan(dataframe)
    data = {
        "read_sync": read,
        "write_sync": write,
        "io_time": {**time_total, **time_io},
    }
    return data, ranks


def extract_data(path: str, args) -> tuple[list, int, dict]:
    """Extracts module from Darshan

    Args:
        path (str): file location
        args (Argparse): optional arguments

    Raises:
        RuntimeError: Not implemented

    Returns:
        _type_: tuple[dataframe,int]
    """
    start = time.time()
    total_time = {}
    with darshan.DarshanReport(path, read_all=True) as report:
        ranks = int(report.metadata["job"]["nprocs"])
        console = Console()
        console.print(f"[cyan]Elapsed time:[/] {time.time()-start:.3f} s")
        start = time.time()
        # dataframe = pd.DataFrame()
        dataframe = []
        # get modules captured
        modules = list(report.modules.keys())
        if isinstance(args, list) or "MPI" in args.dxt_mode.upper():
            if "DXT_MPIIO" in modules:
                dataframe = report.records["DXT_MPIIO"].to_df()
            else:
                console.print("[red]No DXT Module[/]\n[cyan]Trying heatmap[/]")
                dataframe, freq, total_time = extract_heatmap(report, "MPIIO")
                if isinstance(args, list):
                    pass
                elif freq > 0 and "ftio" in args.files[0]:
                    args.freq = freq
                    console.print(f"[cyan]Adjusting sampling freq:[/] {freq:.3e}")

        elif "POSIX" in args.dxt_mode.upper():
            if "DXT_POSIX" in modules:
                dataframe = report.records["DXT_POSIX"].to_df()
            else:
                console.print("[red]No DXT Module[/]\n[cyan]Trying heatmap[/]")
                dataframe, freq, total_time = extract_heatmap(report, "POSIX")
                if freq > 0:
                    args.freq = freq
                    console.print(f"[cyan]Adjusting sampling freq:[/] {freq:.3e}")

        console.print(f"[cyan]Done:[/] {time.time()-start:.3f} s\n")

    return dataframe, ranks, total_time


def extract_heatmap(report, kind: str) -> tuple[list, float, dict]:
    """Extract heatmap to support types

    Args:
        report: darshan report
        kind (str): either MPIIO, POSIX, or STDIO

    Returns:
        list: data
        float: adjusted sampling frequency
        float: total time
    """
    dataframe = []
    total_time = {}
    freq = 0
    for mode in ["read", "write"]:
        heatmap = report.heatmaps[kind].to_df([mode])
        # bins = heatmap._nbins
        bin_width = report.heatmaps[kind]._bin_width_seconds
        freq = 5 / bin_width
        col_name = heatmap.columns.values
        t_sum_rank = 0  # Measures total time
        for rank, row in heatmap.iterrows():
            length = []
            start_time = []
            end_time = []
            for n_bin, value in enumerate(row):
                if value != 0:
                    length.append(value)
                    start_time.append(col_name[n_bin].left)
                    end_time.append(col_name[n_bin].right)
            if length:
                dataframe.append(
                    {
                        "rank": rank,
                        f"{mode}_segments": pd.DataFrame(
                            {
                                "length": length,
                                "start_time": start_time,
                                "end_time": end_time,
                            }
                        ),
                    }
                )
            else:
                dataframe.append({"rank": rank, f"{mode}_segments": pd.DataFrame()})
            t_sum_rank += col_name[-1].right

        total_time["delta_t_agg"] = t_sum_rank

    return dataframe, freq, total_time


def extract_darshan(dataframe: list) -> tuple[dict, dict, dict]:
    """_summary_

    Args:
        df (pd.Dataframe): dataframe contating darshan module

    Returns:
        tuple[dict, dict]:
            1. write dict
            2. read dict
            3. time dict
    """
    write = {
        "number_of_ranks": 0,
        "total_bytes": 0,
        "max_bytes_per_rank": 0,
        "max_bytes_per_phase": 0,
        "max_io_phases_per_rank" : 0,
        "total_io_phases" : 0,
        "bandwidth": {
            "b_rank_sum": [],
            "b_rank_avr": [],
            "t_rank_s": [],
            "t_rank_e": [],
        },
    }
    read = {
        "number_of_ranks": 0,
        "total_bytes": 0,
        "max_bytes_per_rank": 0,
        "max_bytes_per_phase": 0,
        "max_io_phases_per_rank" : 0,
        "total_io_phases" : 0,
        "bandwidth": {
            "b_rank_sum": [],
            "b_rank_avr": [],
            "t_rank_s": [],
            "t_rank_e": [],
        },
    }
    time_sr = 0
    time_sw = 0
    
    
    
    for rank, _ in enumerate(dataframe):
        if (
            "write_segments" in dataframe[rank]
            and not dataframe[rank]["write_segments"].empty
        ):
            bandwidth = (
                dataframe[rank]["write_segments"]["length"]/ (
                    dataframe[rank]["write_segments"]["end_time"]
                    - dataframe[rank]["write_segments"]["start_time"]
                )
            )
            write["bandwidth"]["b_rank_avr"].extend(bandwidth.to_list())
            write["bandwidth"]["b_rank_sum"].extend(bandwidth.to_list())
            write["bandwidth"]["t_rank_e"].extend(
                dataframe[rank]["write_segments"]["end_time"].to_list()
            )
            write["bandwidth"]["t_rank_s"].extend(
                dataframe[rank]["write_segments"]["start_time"].to_list()
            )
            write["number_of_ranks"] = max(write["number_of_ranks"], dataframe[rank]["rank"] + 1)
            time_sw += sum(
                (
                    dataframe[rank]["write_segments"]["end_time"]
                    - dataframe[rank]["write_segments"]["start_time"]
                ).to_list()
            )
            write["total_bytes"] += sum(dataframe[rank]["write_segments"]["length"])
            write["max_bytes_per_rank"] = max(
                write["max_bytes_per_rank"], sum(dataframe[rank]["write_segments"]["length"])
            )
            write["max_bytes_per_phase"] = max(
                write["max_bytes_per_phase"], max(dataframe[rank]["write_segments"]["length"]))
            write["max_io_phases_per_rank"] = max(write["max_io_phases_per_rank"], len(dataframe[rank]["write_segments"]["length"]))
            write["total_io_phases"] += len(dataframe[rank]["write_segments"]["length"])

        if (
            "read_segments" in dataframe[rank]
            and not dataframe[rank]["read_segments"].empty
        ):
            bandwidth = (
                dataframe[rank]["read_segments"]["length"] / (
                    dataframe[rank]["read_segments"]["end_time"]
                    - dataframe[rank]["read_segments"]["start_time"]
                )
            )
            read["bandwidth"]["b_rank_avr"].extend(bandwidth.to_list())
            read["bandwidth"]["b_rank_sum"].extend(bandwidth.to_list())
            read["bandwidth"]["t_rank_e"].extend(
                dataframe[rank]["read_segments"]["end_time"].to_list()
            )
            read["bandwidth"]["t_rank_s"].extend(
                dataframe[rank]["read_segments"]["start_time"].to_list()
            )
            read["number_of_ranks"] = max(read["number_of_ranks"], dataframe[rank]["rank"] + 1)
            time_sr += sum(
                (
                    dataframe[rank]["read_segments"]["end_time"]
                    - dataframe[rank]["read_segments"]["start_time"]
                ).to_list()
            )
            read["total_bytes"] += sum(dataframe[rank]["read_segments"]["length"])
            read["max_bytes_per_rank"] = max(
                read["max_bytes_per_rank"], sum(dataframe[rank]["read_segments"]["length"])
            )
            read["max_bytes_per_phase"] = max(
                read["max_bytes_per_phase"], max(dataframe[rank]["read_segments"]["length"]))
            read["max_io_phases_per_rank"] = max(read["max_io_phases_per_rank"], len(dataframe[rank]["read_segments"]["length"]))
            read["total_io_phases"] += len(dataframe[rank]["read_segments"]["length"])

    # total time
    time = {
        "delta_t_sw": time_sw,
        "delta_t_sr": time_sr,
        "delta_t_awa": 0,
        "delta_t_awr": 0,
        "delta_t_aw_lost": 0,
        "delta_t_ara": 0,
        "delta_t_arr": 0,
        "delta_t_ar_lost": 0,
        "delta_t_overhead": 0,
        "delta_t_agg_io": time_sw + time_sr,
    }

    return write, read, time


def main(args) -> None:
    """Pass varibales and call main_core. The extraction of the traces
    and the parsing of the arguments is done in this function.
    """
    file = args[1]
    data, _ = extract(file, args[1:])
    print(data)


if __name__ == "__main__":
    main(sys.argv)



from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
from ftio.freq.freq_plot_core import convert_and_plot
from ftio.api.metric_proxy.parse_proxy import parse
from ftio.post.processing import label_phases
from ftio.freq.helper import MyConsole

CONSOLE = MyConsole()
CONSOLE.set(True)

# ---------------------------------
# Modification Area
# ---------------------------------
b, t = parse("/d/sim/metric_proxy/traces/Mixed_1x8_5.json", "total___mpi___size_total")
# b, t = parse("/d/sim/metric_proxy/traces/Mixed_1x8_5.json", "mpi___size___mpi_allgather")
ranks = 32

# command line arguments
argv = ["-e", "plotly"]  # ["-e", "no"] to disable the plot
argv.extend(["-n","15"]) # finds up to n frequencies. Comment this out to go back to the default version
# ---------------------------------

# set up data
data = {"time": t, "bandwidth": b, "total_bytes": 0, "ranks": ranks}

# parse args
args = parse_args(argv, "ftio")

# perform prediction
prediction, dfs = core([data], args)

# plot and print info
convert_and_plot(data, dfs, args)
display_prediction("ftio", prediction)

# Post processing
if prediction and len(prediction["dominant_freq"]) != 0:
    phases, time = label_phases(prediction, args, b, t)
    CONSOLE.print(
        f"[cyan]Phases[/]: {phases}\n[cyan]Start time[/]: {time['t_s']}\n[cyan]End time[/]: {time['t_e']}\n"
    )



pattern = {
    "avg_thruput_mib": r"avg_thruput_mib:\s+\[([\d.\d,\s]+)\]",
    "end_t_micro": r"end_t_micro:\s+\[([\d,\s]+)\]",
    "start_t_micro": r"start_t_micro:\s+\[([\d,\s]+)\]",
    "total_bytes": r"total_bytes:\s+(\d+)",
    }

# Define map according to sample.py class, along with the scale if any:
# ftio_field: ("custom_name", scale)
# ftio unit are default in bytes, b/s, ...
# scale applies ftio_field = custom_name*scale
translate = {
    "bandwidth": {
        "b_rank_avr": ("avg_thruput_mib",1.07*1e+6),
        "t_rank_e": ("end_t_micro", 1e-3),
        "t_rank_s": ("start_t_micro", 1e-3)
        },
    "total_bytes": "total_bytes",
    }



from dash_extensions.enrich import DashProxy

from ftio.plot.dash_files.callback_files.io_mode_callbacks import get_io_mode_specific_callbacks
from ftio.plot.dash_files.data_source import get_data_source


def get_callbacks(app: DashProxy, plot_core, io_modes: list[str]) -> None:
    """Is responsible for defining all callbacks for the dash app.

    Args:
        app (DashProxy): The dash app
        plot_core (plot_core): The core component for plotting.
        io_modes (list[str]): All possible io_modes (according to the app).
    """
    for io_mode in io_modes:
        data_source = get_data_source(plot_core, io_mode)
        get_io_mode_specific_callbacks(app, data_source)



import json
from ftio.parse.simrun import Simrun
from ftio.freq.helper import MyConsole
from ftio.parse.helper import match_mode, detect_source

class ParseJson:
    def __init__(self, path):
        self.path = path
        if self.path[-1] == "/":
            self.path = self.path[:-1]

    def to_simrun(self, args, index = 0):
        """Convert to Simrun class
        Args:
            ars (argparse): command line arguments
            index: file index in case several files are passed
        Returns:
            Simrun: Simrun object
        """
        file = self.path
        with open(file, "rt") as current_file:
            data = json.load(current_file)

        source = detect_source(data, args)
        if 'tmio' in source:
            pass
        else:
            data, args = self.adjust(data, args)

        return Simrun(data, "json", file, args, index)


    def adjust(self, data:dict, args):
        # check for mode
        fields = list(data.keys())
        if any( x in fields for x in ['read_sync', 'read_async', 'write_sync', 'write_async']):
            args.source = 'custom'
            if args.mode and match_mode(args.mode) in data:
                self.add_missing_fields(data,match_mode(args.mode))
            else:
                CONSOLE = MyConsole()
                CONSOLE.info(f"[yellow]Warning: [/] Mode [yellow]{args.mode}[/] does not exist in trace")
                for x in ['read_sync', 'read_async', 'read_async', 'write_async']:
                    if x in fields:
                        args.mode = x
                        CONSOLE.info(f"[yellow]Warning: [/] Adjusting mode to [yellow]{args.mode}[/]")
                        self.add_missing_fields(data,match_mode(args.mode))
                        break

        elif  'bandwidth' in data:
            data = {"read_sync":{
                    "total_bytes": 0,
                    "number_of_ranks": 0 ,
                    "bandwidth": data['bandwidth']
                    }}
            args.mode = "read_sync"
            args.source = 'custom'
        elif "b_rank" in data:
            data = {"read_sync":{
                    "total_bytes": 0,
                    "number_of_ranks": 0 ,
                    "bandwidth": 
                        {
                        "b_rank_avr": data["b_rank_avr"],
                        "t_rank_s":  data["t_rank_s"],
                        "t_rank_e": data["t_rank_e"]
                    }
                    }}
            args.mode = "read_sync"
            args.source = 'custom'
        elif "b_overlap" in data:
            data = {"read_sync":{
                    "total_bytes": 0,
                    "number_of_ranks": 0 ,
                    "bandwidth": 
                        {
                        "b_overlap_avr": data["b_overlap"],
                        "t_overlap":  data["t_overlap"],
                    }
                    }}
            args.mode = "read_sync"
            args.source = 'custom'
        else:
            raise ValueError("Mode not found")

        return data, args

    def add_missing_fields(self,data,mode):
        fields = ["total_bytes", "number_of_ranks"]
        for field in fields:
            if field not in data[mode]:
                data[mode][field] = 0




"""
Parses recorder object to simrun
This function can be also executed as a standalone. Just call:
> python3 darshan_reader.py FILE

Returns:
    list[dict]: _description_
"""
import os
import sys 

def extract(path, args)-> tuple[dict,int]:
    """exrtracts Darshan file and generates dictionary with relevent keys

    Args:
        path (str): filename
        args (Argparse): optional arguments

    Returns:
        tuple[dict,int]: 
            1. dictionary with relevant files
            2. number of ranks
    """
    data, ranks = extract_data(path, args)
    write, read = extract_recorder(data, ranks)
    data = {'read_sync':  read, 'write_sync': write}
    return data,ranks


def extract_data(path:str, args)-> tuple[list, int]:
    """Extract relevent fields from all recorder files

    Args:
        path (str): path of file 
        args (Argparse): _description_

    Returns:
        tuple[list, int]: 
        1. list of data contating read and write information
        2. number of ranks 
    """
    data = []
    rank = 0
    for root, _ , files in os.walk(path):
        for file in sorted(files,key=len):
            file = os.path.join(root, file)
            current_file = open(file, 'r')
            current_file = current_file.readlines()
            # data.extend([k for k in f if 'MPI_File_w' in k or 'MPI_File_r' in k])
            data.extend([k for k in current_file if ' write ' in k or ' read ' in k])
        rank = max([int(x.replace(".txt", "")) for x in files])+1
        break #no recursive walk
    return data, rank


def extract_recorder(data:list, ranks:int) -> tuple[dict, dict]:
    """Extract recorder traces

    Args:
        data (array): simulation data

    Returns:
        w,r: returns two dicts containing the data
    """
    # FIXME: This needs to ne b_ind t_ind_s and t_ind_e, however required ind overlaping alorithm in bandiwdth.py 
    write ={'number_of_ranks': ranks, 'bandwidth':{'b_rank_sum': [], 'b_rank_avr': [] ,'t_rank_s':[], 't_rank_e':[] }}
    read = {'number_of_ranks': ranks, 'bandwidth':{'b_rank_sum': [], 'b_rank_avr': [] ,'t_rank_s':[], 't_rank_e':[] }}
    for line in data:
        if 'write' in line or 'read' in line:
            s_line  = line.find(' ')            
            t_start = float(line[:s_line])
            t_end = float(line[s_line+1:s_line+line[s_line+1:].find(' ')+1])
            #? For MPI oly
            # b  = line.rfind('%p')
            # b = int(line[b+3: b+3+line[b+3:].find(' ')])
            #? write
            b_part  = line.rfind(')')
            b_part = int(line[line[:b_part-1].rfind(' ')+1:b_part-1])

            b_part = b_part/(t_end-t_start) if t_end-t_start != 0 else 0 #B/s
            #? Assign
            if 'write' in line:
                write['bandwidth']['t_rank_s'].append(t_start)
                write['bandwidth']['t_rank_e'].append(t_end)
                write['bandwidth']['b_rank_avr'].append(b_part)
                write['bandwidth']['b_rank_sum'].append(b_part)
            elif 'read' in line:
                read['bandwidth']['t_rank_s'].append(t_start)
                read['bandwidth']['t_rank_e'].append(t_end)
                read['bandwidth']['b_rank_avr'].append(b_part)
                read['bandwidth']['b_rank_sum'].append(b_part)

    if not read['bandwidth']['t_rank_s']:
        read = {'bandwidth':[]}
    if not write['bandwidth']['t_rank_s']:
        write = {'bandwidth':[]}
    return write, read


def main(args) -> None:
    """Pass varibales and call main_core. The extraction of the traces
    and the parsing of the arguments is done in this function.
    """
    file = args[1]
    dataframe,_ = extract(file,args[1:])
    print(dataframe)


if __name__ == "__main__":
    main(sys.argv)



"""IDs for dash app components"""

import ftio.plot.dash_files.constants.io_mode as io_mode

# Button
BUTTON_SHOW: str = "button-show"

# Checklist
CHECKLIST_IO_MODE: str = "checklist-io-mode"

# Div
DIV_SYNC_READ: str = "div-sync-read"
DIV_SYNC_WRITE: str = "div-sync-write"
DIV_ASNYC_READ: str = "div-async-read"
DIV_ASYNC_WRITE: str = "div-async-write"
DIV_IO_TIME: str = "div-io-time"

DIV_IO_BY_IO_MODE: dict[str, str] = {
    io_mode.SYNC_READ: DIV_SYNC_READ,
    io_mode.SYNC_WRITE: DIV_SYNC_WRITE,
    io_mode.ASYNC_READ: DIV_ASNYC_READ,
    io_mode.ASYNC_WRITE: DIV_ASYNC_WRITE,
    io_mode.TIME: DIV_IO_TIME,
}

# Dropdown
DROPDOWN_FILE: str = "dropdown-file"

# Store
STORE_FIGURES_ASYNC_READ: str = "store-figures-" + io_mode.ASYNC_READ
STORE_FIGURES_ASYNC_WRITE: str = "store-figures-" + io_mode.ASYNC_WRITE
STORE_FIGURES_SYNC_READ: str = "store-figures-" + io_mode.SYNC_READ
STORE_FIGURES_SYNC_WRITE: str = "store-figures-" + io_mode.SYNC_WRITE
STORE_FIGURES_TIME: str = "store-figures-" + io_mode.TIME

STORE_FIGURES_BY_IO_MODE: dict[str, str] = {
    io_mode.ASYNC_READ: STORE_FIGURES_ASYNC_READ,
    io_mode.ASYNC_WRITE: STORE_FIGURES_ASYNC_WRITE,
    io_mode.SYNC_READ: STORE_FIGURES_SYNC_READ,
    io_mode.SYNC_WRITE: STORE_FIGURES_SYNC_WRITE,
    io_mode.TIME: STORE_FIGURES_TIME,
}

# Type
TYPE_DYNAMIC_GRAPH: str = "dynamic-graph"

TYPE_DYNAMIC_UPDATER_ASYNC_READ: str = "dynamic-updater-" + io_mode.ASYNC_READ
TYPE_DYNAMIC_UPDATER_ASYNC_WRITE: str = "dynamic-updater-" + io_mode.ASYNC_WRITE
TYPE_DYNAMIC_UPDATER_SYNC_READ: str = "dynamic-updater-" + io_mode.SYNC_READ
TYPE_DYNAMIC_UPDATER_SYNC_WRITE: str = "dynamic-updater-" + io_mode.SYNC_WRITE
TYPE_DYNAMIC_UPDATER_TIME: str = "dynamic-updater-" + io_mode.TIME

TYPE_DYNAMIC_UPDATER_BY_IO_MODE: dict[str, str] = {
    io_mode.ASYNC_READ: TYPE_DYNAMIC_UPDATER_ASYNC_READ,
    io_mode.ASYNC_WRITE: TYPE_DYNAMIC_UPDATER_ASYNC_WRITE,
    io_mode.SYNC_READ: TYPE_DYNAMIC_UPDATER_SYNC_READ,
    io_mode.SYNC_WRITE: TYPE_DYNAMIC_UPDATER_SYNC_WRITE,
    io_mode.TIME: TYPE_DYNAMIC_UPDATER_TIME,
}



import numpy as np
import json


def parse(file_path, match="proxy_component_critical_temperature_celcius")-> tuple[np.ndarray,np.ndarray]:
    b_out = np.array([])
    t_out = np.array([])
    try:
        with open(file_path, 'r') as json_file:
            json_data = json.load(json_file)
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
        return b_out,t_out
    except json.JSONDecodeError:
        print(f"Error: Unable to decode JSON from file '{file_path}'. Check if the file is valid JSON.")
        return b_out,t_out

    b_out,t_out = extract(json_data, match)

    if len(b_out) == 0:
        print("No match found. Exciting\n")
        exit(0)
    return b_out,t_out


def extract(json_data, match):
    b_out = np.array([])
    t_out  = np.array([])
    for key, value in json_data.items():
        if isinstance(value, dict):
            b_out,t_out = extract(value, match)
            if len(b_out) > 0:
                break
        else:
            if match == key:
                print(f"matched {key}")
                x = np.array(value)
                t_out = x[:,0]
                b_out = x[:,1]
                #reduce to derivative
                if "deriv" not in key:
                    print("removing aggregation")
                    b_shifted = b_out[:-1]
                    b_shifted = np.insert(b_shifted,0,0)
                    b_out =  b_out - b_shifted
                    break
    return b_out,t_out




from __future__ import annotations
import os
import fcntl
import signal
from time import sleep
from rich.console import Console
from ftio.prediction.async_process import join_procs
from time import sleep

CONSOLE = Console()


def monitor(name: str, _cached_stamp: str, procs: list = []) ->  tuple[str, list]:
    """Monitors a file for change and can optionally join processes in the mean time

    Args:
        name (str): filename
        _cached_stamp (str): time stamp since last time
        procs (list, optional): List of proccess to join. Defaults to [].

    Returns:
        str: _description_
    """
    return monitor_stat(name, _cached_stamp, procs)
    # try:
    #     return monitor_stat(name, _cached_stamp, procs)
    # finally:
    #     return monitor_fcntl(name, _cached_stamp, procs)


#! Method 1
def monitor_stat(name: str, _cached_stamp: str, procs: list) -> tuple[str, list]:
    """Monitors a file for changes

    Args:
        name (str): _description_
        _cached_stamp (str): change time stamp
        procs: list or process

    Returns:
        str: _description_
    """
    if _cached_stamp == "":
        stream = os.popen(f"stat -c %z {name} 2>&1")
        stamp = stream.read()
        CONSOLE.print(f"[purple][PREDICTOR]:[/] Monitoring file {name}")
        CONSOLE.print(f"[purple][PREDICTOR]:[/] Stamp is {stamp}")
        return stamp, procs
    else:
        sleep(1)
        while True:
            procs = join_procs(procs)
            stream = os.popen(f"stat -c %z {name} 2>&1")
            stamp = stream.read()
            if stamp != _cached_stamp:
                CONSOLE.print(
                    f"\n[purple][PREDICTOR]:[/][red bold]Stamp changed[/] to {stamp}"
                )
                sleep(0.2)
                return stamp, procs

def monitor_list(name: list, n_buffers, _cached_stamp: dict={}, procs: list = []) -> tuple[dict, list]:
    """Monitors a file for changes

    Args:
        name (str): _description_
        _cached_stamp (str): change time stamp
        procs: list or process

    Returns:
        str: _description_
    """
    if not _cached_stamp:
        stamp = {}
        for i in name:
            stream = os.popen(f"stat -c %z {i} 2>&1")
            file_stamp = stream.read()
            stamp[i] = file_stamp
            CONSOLE.print(
                f"[purple][PREDICTOR]:[/] Monitoring file {name.index(i)}/{n_buffers} {i}"
                f"[purple][PREDICTOR]:[/] Stamp is {stamp[i]}"
                )
        return stamp, procs
    else:
        sleep(1)
        seen = []
        counter = n_buffers
        while True:
            procs = join_procs(procs)
            for _,i in enumerate(_cached_stamp):
                stream = os.popen(f"stat -c %z {i} 2>&1")
                file_stamp = stream.read()
                if file_stamp != _cached_stamp[i] and  _cached_stamp[i] not in seen:
                    counter = counter - 1
                    seen.append(_cached_stamp[i])
                    CONSOLE.print(
                        f"[purple][PREDICTOR]:[/][red bold]Stamp changed[/] to {file_stamp}"
                        f"[purple][PREDICTOR]:[/] {n_buffers - counter}/{n_buffers} files changed"
                        )

            if counter == 0:
                for i in name:
                    stream = os.popen(f"stat -c %z {i} 2>&1")
                    file_stamp = stream.read()
                    _cached_stamp[i] = file_stamp
                return _cached_stamp, procs
            
            sleep(0.2)

#! Method 2
def monitor_fcntl(name: str, _cached_stamp: str, procs: list) ->  tuple[str, list]:
    if _cached_stamp == 0:
        return "", procs

    w = watcher(name)
    signal.signal(signal.SIGIO, w.handler)
    fd = os.open(w.path, os.O_RDONLY)
    fcntl.fcntl(fd, fcntl.F_SETSIG, 0)
    fcntl.fcntl(
        fd,
        fcntl.F_NOTIFY,
        fcntl.DN_ACCESS
        | fcntl.DN_MODIFY
        | fcntl.DN_CREATE
        | fcntl.DN_DELETE
        | fcntl.DN_RENAME
        | fcntl.DN_ATTRIB
        | fcntl.DN_MULTISHOT,
    )
    while w.flag:
        sleep(0.1)
        procs = join_procs(procs)

    return str(w.stamp),procs


class watcher:
    def __init__(self, name):
        self.flag = True
        if "/" in name:
            index = name.rfind("/")
            self.path = name[:index]
            self.name = name[index + 1 :]
        else:
            self.path = os.getcwd()
            self.name = name
        self.absolute_name = f"{self.path}/{self.name}"
        print(f"Monitoring file {self.absolute_name} in path:{self.path}")
        self.stamp = os.stat(self.absolute_name).st_mtime
        CONSOLE.print(f"[purple][PREDICTOR]:[/] Stamp is {self.stamp}")

    def handler(self, signum, path):
        CONSOLE.print(f"[purple][PREDICTOR]:[/] Folder {path} modified ")
        stamp = os.stat(self.absolute_name).st_mtime
        if self.stamp != stamp:
            self.flag = False
            CONSOLE.print(
                f"\n[purple][PREDICTOR]:[/] [red bold]Stamp changed[/] to {stamp}"
            )



"""
Parses messagpack object to a format similar to jsonl
This function can be also executed as a standalone. Just call:
> python3 msgpack_reader.py FILE

Returns:
    list[dict]: _description_
"""
from __future__ import annotations
import sys
import msgpack



def get_type(arr:list[str]) -> str:
    io_type = ""
    if arr:
        io_type = arr[0]
    return io_type


def convert_to_class(arr: list, io_type: str) -> list:
    if arr:
        if "write" in io_type or "read" in io_type:
            return convert_to_iosample(arr, io_type)
        elif "time" in io_type:
            return convert_to_iotime(arr)
        else:
            raise ValueError(f"Parse error: type is {io_type}")
            


def convert_to_iosample(arr: list, io_type: str) -> list:
    """Brings the data into an structure similar to Sample/Time 
    depending on the type of the I/O (read/write + sync/async | time)

    Args:
        arr (list): metrics
        io_type (str): type of the I/O (read/write + sync/async | time)

    Returns:
        list: list
    """
    if "async" in io_type:
        bandwidth = assign_bandwidth(arr,io_type)
        return [{f"{io_type}_t":assign_metrics(arr,bandwidth[0])},{f"{io_type}_b":assign_metrics(arr,bandwidth[1])}]
    else:
        bandwidth = assign_bandwidth(arr,io_type)
        return [{io_type:assign_metrics(arr,bandwidth[0])}]


def assign_metrics(arr: list, bandwidth:dict) -> dict:
    return {
            "total_bytes": arr[3] , #Bytes
            "max_bytes_per_rank": arr[4],
            "max_bytes_per_phase": arr[5],
            "max_io_phases_per_rank": arr[6],
            "total_io_phases": arr[7],
            "max_io_ops_in_phase": arr[8],
            "max_io_ops_per_rank": arr[9],
            "total_io_ops": arr[10],
            "number_of_ranks": arr[11],
            "total_number_of_ranks": arr[12],
            "bandwidth": bandwidth,
        }

    
def assign_bandwidth(arr: list, io_type: str) -> list[dict]:
    # data = []
    t_start = []
    t_end_act = []
    t_end_req = []
    T_sum = []
    T_avr = []
    B_sum = []
    B_avr = []
    if arr[13]:
        for i in arr[13]:
            # data.append(i[0])
            t_start.append(i[1])
            t_end_act.append(i[2])
            T_sum.append(i[4]) # Bytes/s
            T_avr.append(i[5]) # Bytes/s
            if "async" in io_type:
                t_end_req.append(i[3])
                B_sum.append(i[6]) # Bytes/s
                B_avr.append(i[7]) # Bytes/s
            # n_op.append(i[8])
    if "async" in io_type:
        return [{
            'b_rank_sum': T_sum,
            'b_rank_avr': T_avr,
            't_rank_s':  t_start,
            't_rank_e': t_end_act      
            },{
            'b_rank_sum': B_sum,
            'b_rank_avr': B_avr,
            't_rank_s':  t_start,
            't_rank_e': t_end_req      
            }]
    else:
        return [{
            'b_rank_sum': T_sum,
            'b_rank_avr': T_avr,
            't_rank_s':  t_start,
            't_rank_e': t_end_act      
            }]


def convert_to_iotime(arr: list) -> list:
    tmp = {
        'delta_t_agg' : arr[1],
        'delta_t_sr' : arr[2],
        'delta_t_sw' : arr[3],
        'delta_t_ara' : arr[4],
        'delta_t_arr' : arr[5],
        'delta_t_ar_lost' : arr[6],
        'delta_t_awa' : arr[7],
        'delta_t_awr' : arr[8],
        'delta_t_aw_lost' : arr[9],
        'delta_t_agg_io' : arr[10],
        'delta_t_rank0' : arr[11],
        'delta_t_rank0_app' : arr[12],
        'delta_t_rank0_overhead_post_runtime' : arr[13],
        'delta_t_rank0_overhead_peri_runtime' : arr[14],
        'delta_t_overhead': arr[15],
        'delta_t_overhead_post_runtime': arr[16],
        'delta_t_overhead_peri_runtime': arr[17],
        'delta_t_overhead_dft' : arr[18]
        }
    return [{'io_time':tmp}]


def extract(file:str) -> list[dict]:
    """Extracts the data stored in Darshan file

    Args:
        file (str): file name

    Returns:
        list[dict]: file content
    """
    data = []
    # Open the MessagePack binary file for reading
    with open(file, "rb") as in_file:
        # Read the binary data
        binary_data = in_file.read()

    data = extract_data(binary_data, data)


def extract_data(binary_data, data):
    # Deserialize the MessagePack data into the class instances
    unpacker = msgpack.Unpacker()
    unpacker.feed(binary_data)
    for item in unpacker:
        io_type = get_type(item)
        data.extend(convert_to_class(item, io_type))
    return data

def main(args) -> None:
    """Pass variables and call main_core. The extraction of the traces
    and the parsing of the arguments is done in this function.
    """
    file = args[1]
    extract(file)


if __name__ == "__main__":
    main(sys.argv)



import numpy as np

def scale_metric(metric:str,number:float) -> tuple[str, float]:
    """set unit for the plots

    Args:
        number (np.ndarray): array

    Returns:
        unit (string): unit in GB/s, MB/s, KB/s or B/s
        number: scaled array according to the unit
    """
    unit = metric
    order = 1e-0
    prefix = ""
    if number > 0:
        order = np.log10(number)

        if any(x in metric.lower() for x in ["bytes","b", "bandwidth", "transfer"]):
            if order > 9:
                order = 1e-9
                prefix = "G"
            elif order > 6:
                order = 1e-6
                prefix = "M"
            elif order > 3:
                order = 1e-3
                prefix = "K"
            else:
                order = 1e-0
                prefix =""
        elif any(x in metric.lower() for x in ["time","(s)"]):
            if order > 0:
                order = 1e-0
                prefix = ""
            elif order > -3:
                order = 1e-3
                prefix = "μ"
            else:
                order = np.nan
        else:
            return metric, 1e-0

        if not np.isnan(order):
            if any(x in metric.lower() for x in ["(b/s)","(bytes/s)", "(bytes/second)", "(b/second)", "bandwidth"]):
                unit = f"Bandwidth ({prefix}B/s)"
            elif any(x in metric.lower() for x in ["(bytes)","(b)"]):
                unit = f"Size ({prefix}B)"
            elif any(x in metric.lower() for x in ["(s)","(second)", "time"]):
                unit = f"Time ({prefix}s)"
            else:
                unit = "UNKOWN"
        else:
            unit = "UNKOWN"
            order = 1e-0

    return unit, order


def match_modes(mode):
    mode = mode.lower()
    if isinstance(mode, list):
        for i in range(0, len(mode)):
            mode[i] = match_mode(mode[i])
    else:
        mode = [match_mode(mode)]
    return mode


def match_mode( mode:str) -> str:
    out  = ''
    if "w" in mode:
        out = "write"
    else:
        out = "read"
        
    if "async" in mode:
        out = out + "_async"
    else:
        out = out + "_sync"

    return out 

def detect_source(data:dict, args) -> str:
    if 'tmio' in args.source.lower() or 'custom' in args.source.lower():
        return args.source.lower()
    else: #autodetect
        tmio_fields = ['read_sync', 'read_async_t', 'read_async_b', 'write_async_t', 'io_time']
        if all(fields in data for fields in tmio_fields):
            return 'tmio'
        else:
            return 'unspecified'



"""Handles the importing of modules
"""

import os
import datetime
import pandas as pd
import numpy as np
from rich.panel import Panel
from rich.console import Console
from rich.text import Text
from ftio import __version__
from ftio.parse.parse_json import ParseJson
from ftio.parse.parse_jsonl import ParseJsonl
from ftio.parse.parse_recorder import ParseRecorder
from ftio.parse.parse_darshan import ParseDarshan
from ftio.parse.parse_msgpack import ParseMsgpack
from ftio.parse.parse_txt import ParseTxt
from ftio.parse.parse_custom import ParseCustom
from ftio.parse.parse_zmq import ParseZmq
from ftio.parse.args import parse_args


class Scales:
    """load the data. Supports single files (json, jsonl, darshan) or folders (+ recorder)"""

    def __init__(self, argv, msg=None):
        self.Print_info(argv[0])
        self.render = ""
        self.plot_mode = ""
        self.mode = ""
        self.ts = -1
        self.threaded = ""
        self.zoom = -1
        self.same_path = False
        self.names = []
        self.msg = msg
        self.s = []

        # save call
        self.save_call(argv)

        # Parse arguments
        self.args = parse_args(argv)

        if "zmq" in self.args and self.args.zmq:
            self.s.append(ParseZmq(self.msg).to_simrun(self.args, 0))
            self.n = 1
        else:
            self.load_setup()

    def load_setup(self)-> None:
        if isinstance(self.args.files, list):
            if len(self.args.files) <= 1:
                self.paths = ["."]
            else:
                self.paths = []
                for i in range(1, len(self.args.files)):
                    self.paths.append(str(self.args.files[i]))
        else:
            self.paths = [self.args.files]

        self.Check_Same_Path()
        console = Console()
        for path in self.paths:
            #! load folders
            # Recorder folder
            if "_text" in path[-5:]:
                console.print(
                    f"\n[cyan]Loading Recorder folder({self.paths.index(path) + 1},{len(self.paths)}):[/] {path}"
                )
                run = ParseRecorder(path).to_simrun(self.args)
                self.s.append(run)

            # Folder
            elif os.path.isdir(path):
                console.print(
                    f"\n[cyan]Loading folder({self.paths.index(path) + 1},{len(self.paths)}):[/] {path}"
                )
                if path[-1] == "/":
                    path = path[:-1]

                for root, _, files in os.walk(path):
                    # remove unneeded folders
                    if "io_results" in root or "exported_images" in root:
                        console.print(f"[yellow]Skipping folder:  {root}[/]")
                        continue

                    # remove unneeded files
                    skip_files = ["scale.jsonl",".call.txt"]
                    for unwanted in skip_files:
                        if unwanted in files:
                            console.print(
                                f"[yellow]Skipping file: {root}/{unwanted}[/]"
                            )
                            files.remove(unwanted)

                    # sort the files
                    sorted_files = sorted(files, key=len)

                    for file in sorted_files:
                        if any(
                            ext in file for ext in ["json", "darshan", "msgpack", "txt"]
                        ):
                            file_path = os.path.join(root, file)
                            # Limit the number of ranks to consider if self.limit is defined
                            try:
                                if (
                                    self.args.limit > 0
                                    and get_rank(file) >= self.args.limit
                                ):
                                    console.print(f"[yellow]Skipping file: {file}[/]")
                                    continue
                            except Exception as error:
                                console.print(
                                    f"[red]Something went wrong with the limit. Error is {error}[/]"
                                )
                            self.names.append(root[root.rfind("/") + 1 :])
                            console.print(f"[cyan]Current file:[/] {file}")
                            self.load_file(file_path, self.paths.index(path))
                    break  # no reclusive walk

            # Compare Several files
            elif not self.same_path and ".json" in path[-6:]:
                self.names.append(path)
                console.print(f"[cyan]Current file:[/] {path}")
                self.load_file(path, self.paths.index(path))

            # Single file
            else:
                self.names.append(get_filename(path))
                console.print(f"[cyan]Current file:[/] {path}\n")
                self.load_file(path)

        # print("--------------------------------------------\n")
        self.n = len(self.s)
        if self.names:
            names = self.names
            self.names = []
            for i in names:
                if i not in self.names:
                    self.names.append(i)

    def load_file(self, file_path: str, file_index=0) -> None:
        """Load file content into an Simrun object

        Args:
            file_path (str): filename + absolute path
        """
        check_open(file_path)
        if self.args.custom_file:
            run = ParseCustom(file_path).to_simrun(self.args, file_index)
        elif ".json" in file_path[-5:]:
            run = ParseJson(file_path).to_simrun(self.args, file_index)
        elif ".jsonl" in file_path[-6:]:
            run = ParseJsonl(file_path).to_simrun(self.args, file_index)
        elif "darshan" in file_path[-10:]:
            run = ParseDarshan(file_path).to_simrun(self.args, file_index)
        elif "msgpack" in file_path[-10:]:
            run = ParseMsgpack(file_path).to_simrun(self.args, file_index)
        elif "txt" in file_path[-10:]:
            run = ParseTxt(file_path).to_simrun(self.args, file_index)
        self.s.append(run)

    def save_call(self, argv):
        """save the call as a hidden file"""
        self.call = ""
        for i in argv:
            self.call = self.call + " " + i
        f = open("%s/.call.txt" % (os.getcwd()), "a")
        f.write(
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M")
            + " :"
            + self.call
            + "\n\n"
        )
        f.close()

    def Print_info(self, text: str) -> None:
        name = text[text.rfind("/") + 1 :].capitalize()
        console = Console()
        title = Panel(
            Text(name, justify="center"),
            style="bold white on cyan",
            border_style="white",
            title_align="left",
        )
        text = "\n[cyan]Author:[/]  Ahmad Tarraf\n"
        text += f"[cyan]Date:[/]    {str(datetime.date.today())}\n"
        text += f"[cyan]Version:[/]  {__version__}\n"
        text += "[cyan]License:[/]  BSD\n"
        # console.print(Panel(Group(title,text), style="white", border_style='blue'))
        console.print(title)
        console.print(text)

    def Check_Same_Path(self):
        if len(self.paths) > 1:
            self.same_path = True
            same_path = ""
            for i in self.paths:
                if self.paths.index(i) == 0:
                    same_path = i[: i.rfind("/")]
                else:
                    if same_path != i[: i.rfind("/")]:
                        self.same_path = False
                        break
        else:
            self.same_path = True

    #! ----------------------- Pandas dataset functions ------------------------------
    # **********************************************************************
    # *                       1. get_data
    # **********************************************************************
    def get_data(self):
        self.df_rst = self.get_data_io("read_sync")
        self.df_wst = self.get_data_io("write_sync")
        self.df_rat = self.get_data_io("read_async_t")
        self.df_rab = self.get_data_io("read_async_b")
        self.df_wat = self.get_data_io("write_async_t")
        self.df_wab = self.get_data_io("write_async_b")
        self.df_time = self.get_data_time("io_time")

        # df = pd.concat([df,self.get_data_core('write_async_b')])

    # **********************************************************************
    # *                       2. get_data_io
    # **********************************************************************
    def get_data_io(self, io_mode="read_sync"):
        """Extract data from the file(s) and gathers in dataframes.
        The fields name are store in "name_[level]" and their values are stored
        in "data_[level]". There are 4 levels provided:
        (1) Application level (overlap or rank metrics): [..]_rank_ovr
        (2) rank level (sum/average of I/O requests): [..]_rank
        (3) high precision rank level (overlap of I/O requests): [..]_ind_ovr
        (4) I/O request level (lowest level): [..]_rank_over

        if the 'ind' flag is not provided, level (3) and (4) are skipped as they are
        expensive to calculate.


        Args:
            io_mode (str, optional): Can be read or write and
            sync or async (required [b] or actual [t]). Supported modes are:
            "read_sync", "write_sync", "read_async_t", "read_async_b", "write_async_t",
            and "write_async_b".Defaults to "read_sync".

        Returns:
            tuple[pd.DataFrame,pd.DataFrame,
            pd.DataFrame, pd.DataFrame, pd.DataFrame,]: Five dataframes are
            returned. The first one contains metrics like total bytes transferred, number
            of phases, etc.. The next 4 dataframes contain the I/O data at the 4
            levels explained above
        """
        name = ""
        data_metrics = np.array([])
        for i in range(0, self.n):
            value = getattr(self.s[i], io_mode)
            data = value.get_data()
            if i == 0:
                name = data[0]
                name_rank_ovr = data[2]
                name_rank = data[4]
                name_ind_ovr = data[6]
                name_ind = data[8]
                data_metrics = np.array(data[1])
                data_rank_ovr = np.array(data[3])
                data_rank = np.array(data[5])
                data_ind_ovr = np.array(data[7])
                data_ind = np.array(data[9])

            else:
                data_metrics = np.vstack((data_metrics, data[1]))
                data_rank_ovr = np.concatenate((data_rank_ovr, data[3]), axis=1)
                data_rank = np.concatenate((data_rank, data[5]), axis=1)
                data_ind_ovr = np.concatenate((data_ind_ovr, data[7]), axis=1)
                data_ind = np.concatenate((data_ind, data[9]), axis=1)
            # data.append(d0)
        # check if there is data
        if data_metrics.size == 0:
            raise RuntimeError(
                f"The mode {self.args.mode} contains no values\nChange the mode by using the -m argument."
            )

        df0 = pd.DataFrame(data_metrics, columns=name)
        df0 = df0.sort_values(by=["number_of_ranks"])
        df1 = pd.DataFrame(data_rank_ovr.transpose(), columns=name_rank_ovr)
        df1 = df1.astype({"number_of_ranks": "int"})
        df2 = pd.DataFrame(data_rank.transpose(), columns=name_rank)
        df3 = pd.DataFrame(data_ind_ovr.transpose(), columns=name_ind_ovr)
        df4 = pd.DataFrame(data_ind.transpose(), columns=name_ind)

        return df0, df1, df2, df3, df4

    # **********************************************************************
    # *                       3. get_data_time
    # **********************************************************************
    def get_data_time(self, io_mode="io_time"):
        data = []
        name = ""
        for i in range(0, self.n):
            value = getattr(self.s[i], io_mode)
            if i == 0:
                name, d0 = value.get_data()  # Time function
            else:
                _, d0 = value.get_data()

            data.append(d0)
        df0 = pd.DataFrame(data, columns=name)
        df0 = df0.sort_values(by=["number_of_ranks"])
        return df0


def check_open(file: str) -> None:
    """Checks that the file is accessible

    Args:
        file (str): filename
    """
    if os.path.isfile(file):
        pass
    else:
        console = Console()
        console.print(
            f"[red]--- Error  --- [/]\n"
            f"[red]-> Could not open file [b]{file}[/b]. \n[/]"
            f"[yellow]Make sure [b]{file}[/b] exists in [b]{os.getcwd()}[/b]. \n\n[/]"
        )
        exit()


def get_rank(name: str) -> int:
    """Get the number of ranks from the name of the file

    Args:
        name (str): name of file

    Returns:
        int: number of ranks
    """
    if isinstance(name, int):
        return name
    else:
        start = name.rfind("/")
        end = max(name.rfind(".json"), name.rfind(".darshan"), name.rfind(".msgpack"))
        rank = name[start + 1 : end]
        strs = ["_", "-", " "]
        if any(x in rank for x in strs):
            for x in strs:
                if x in rank:
                    end = rank.rfind(x)
                    rank = rank[:end]
        return int(rank)


def get_filename(path: str) -> str:
    """Reutrns filename from absolute path

    Args:
        path (str): absolute path to file (including name)

    Returns:
        str: filename
    """
    tmp = path.rfind("/")
    out = path[tmp + 1 :] if tmp > 0 else path
    return out
